import { BlockPermutation, system, Vector, world, _PI_180_ } from "../lib/Util";

world.beforeEvents.worldInitialize.subscribe(e => {
    e.blockComponentRegistry.registerCustomComponent('mship:cannon_block', {
        beforeOnPlayerPlace: e => {
            let adir = e.permutationToPlace.getState("minecraft:cardinal_direction"), a = e.block, b = a[adir]();
            const bdir = { "north": "south", "south": "north", "west": "east", "east": "west" }[adir];
            const angle = { "west": -90, "north": 0, "east": 90, "south": 180 }[adir]
            if (b.typeId !== "minecraft:air") {
                b = e.block;
                a = b[bdir]();
                e.cancel = a.typeId !== "minecraft:air";
            }
            if (e.cancel) return;
            a.dimension.runCommandAsync(`summon mship:cannon ${a.x + 0.5} ${a.y} ${a.z + 0.5} ${angle} 0`);
            system.run(_ => {
                a.dimension.setBlockPermutation(a, BlockPermutation.resolve("mship:cannon_block", { "minecraft:cardinal_direction": adir, "mship:part": 0, "mship:set": 1 }));
                b.dimension.setBlockPermutation(b, BlockPermutation.resolve("mship:cannon_block", { "minecraft:cardinal_direction": bdir, "mship:part": 1, "mship:set": 1 }));
            });
        },
        onPlace: e => {
            system.run(_ => {
                let a = e.dimension.getBlock(e.block);
                if (a.typeId !== "mship:cannon") return;
                let adir = a.getState("minecraft:cardinal_direction"), b = a[adir]();
                const bdir = { "north": "south", "south": "north", "west": "east", "east": "west" }[adir];
                if (b.typeId !== "minecraft:air") {
                    b = e.block;
                    a = b[bdir]();
                    if (a.typeId !== "minecraft:air") return b.dimension.runCommandAsync(`setblock ${b.x} ${b.y} ${b.z} air destroy`);
                }
                a.dimension.setBlockPermutation(a, BlockPermutation.resolve("mship:cannon_block", { "minecraft:cardinal_direction": adir, "mship:part": 0, "mship:set": 1 }));
                a.dimension.setBlockPermutation(b, BlockPermutation.resolve("mship:cannon_block", { "minecraft:cardinal_direction": bdir, "mship:part": 1, "mship:set": 1 }));
            });
        }
    });
});

world.beforeEvents.worldInitialize.subscribe(e => {
    e.blockComponentRegistry.registerCustomComponent('mship:cannon', {
        onPlayerInteract: e => {
            let targetBlock;
            if (e.block.getState("mship:part") == 0) targetBlock = e.block;
            else targetBlock = e.block[e.block.getState("minecraft:cardinal_direction")]();
            e.block.dimension.getEntitiesAtBlockLocation(targetBlock).forEach(entity => {
                if (entity.hasFamily("cannon")) {
                    if (entity.getRider("minecraft:player")) return;

                    let item = e.player.getItem();
                    if (item?.typeId === "mship:cannon_ball") {
                        if (!entity.getComponent("minecraft:is_ignited")) {
                            if (e.player.getGameMode() === "survival") e.player.setItem(item.setAmount(-1));
                            entity.triggerEvent("mship:fire");
                        }
                        return;
                    }
                    e.player.runCommandAsync(`tp @s ~~~ 0 ${entity.getProperty("mship:pitch") + 4}`).then(_ => entity.addRider(e.player));
                }
            });
        },
        onPlayerDestroy: e => {
            let dir = e.destroyedBlockPermutation.getState("minecraft:cardinal_direction"), other = e.block[dir](), destroy = false;
            if (other && other.hasTag("cannon")) {
                if (e.destroyedBlockPermutation.getState("mship:part") == 1)
                    other.dimension.runCommandAsync(`setblock ${other.x} ${other.y} ${other.z} air destroy`);
                else other.setType("minecraft:air");
            }
        },
        onTick: e => {
            if (e.block.typeId != "mship:cannon_block") return;
            let dir = e.block.getState("minecraft:cardinal_direction"), other = e.block[dir]();
            if (!other || (other && !other.hasTag("cannon"))) e.block.dimension.runCommandAsync(`setblock ${e.block.x} ${e.block.y} ${e.block.z} air destroy`);
            if (e.block.getState("mship:part") == 0) {
                let a = { "west": -90, "north": 0, "east": 90, "south": 180 }[dir], hasCannon = false;
                e.block.dimension.getEntitiesAtBlockLocation(e.block).forEach(entity => {
                    if (entity.hasFamily("cannon")) {
                        if (hasCannon) entity.remove(); // remove duplicates
                        else hasCannon = true;
                    }
                });
                if (!hasCannon) e.block.dimension.runCommandAsync(`summon mship:cannon ${e.block.x + 0.5} ${e.block.y} ${e.block.z + 0.5} ${a} 0`);
            }
        }
    });
});

world.afterEvents.dataDrivenEntityTrigger.subscribe(e => {
    try {
        let { entity, eventId } = e;
        if (!(entity && entity.isValid())) return;
        if (entity.typeId === "mship:cannon") {
            if (eventId == "mship:shoot") {
                const dist = 2.2, vel = 2, id = "mship:cannonball_projectile";
                const r = { x: -entity.getProperty("mship:pitch") * _PI_180_, y: -entity.getRotation().y * _PI_180_ };
                const dir = new Vector(Math.sin(r.y) * Math.cos(r.x), Math.sin(r.x), Math.cos(r.y) * Math.cos(r.x));
                const cannonball = entity.dimension.spawnEntity(id, entity.v3().above().add(dir.clone().mult(dist)));
                cannonball.getComponent("minecraft:projectile").shoot(dir.clone().mult(vel), { uncertainty: 0.1 });
                cannonball.addTag("cannon");
            }
        }
    } catch (err) { throw err; }
});

export default {};