import { BlockPermutation, system, Vector, world, _PI_180_ } from "../lib/Util";

function checkForSurroundingBlocks(block){
    for (let x = -1; x < 2; x++) {
        for (let z = -1; z < 2; z++) {
            if(x == 0 && z == 0) continue;
            const b = block.offset({ x, y: 0, z });
            if (b.typeId !== "minecraft:air") {
                if (b.typeId.includes("cannon")) return true;
                if (b.typeId.includes("helm")) return true;
            }
        }
    }
    return false;
}

world.beforeEvents.worldInitialize.subscribe(e => {
    e.blockComponentRegistry.registerCustomComponent('mship:swivel_cannon', {
        beforeOnPlayerPlace: e => {
            if (checkForSurroundingBlocks(e.block)) return e.cancel = true;
        },
        onPlace: e => {
            if (checkForSurroundingBlocks(e.block)) return e.block.dimension.runCommandAsync(`setblock ${e.block.x} ${e.block.y} ${e.block.z} air destroy`);
            let adir = e.block.getState("minecraft:cardinal_direction"), a = e.block;
            const angle = { "west": -90, "north": 0, "east": 90, "south": 180 }[adir]
            a.dimension.runCommandAsync(`summon mship:swivel_cannon ${a.x + 0.5} ${a.y} ${a.z + 0.5} ${angle} 0`);
            e.block.editState({ "mship:set": 1 });
        },
        onPlayerInteract: e => {
            e.block.dimension.getEntitiesAtBlockLocation(e.block).forEach(entity => {
                if (entity.hasFamily("cannon")) {
                    if (entity.getRider("minecraft:player")) return;

                    let item = e.player.getItem();
                    if (item?.typeId === "mship:cannon_ball") {
                        if (!entity.getComponent("minecraft:is_ignited")) {
                            if (e.player.getGameMode() === "survival") e.player.setItem(item.setAmount(-1));
                            entity.triggerEvent("mship:fire");
                        }
                        return;
                    }
                    e.player.runCommandAsync(`tp @s ~~~ ${entity.getRotation().y} ${entity.getProperty("mship:pitch") + 4}`).then(_ => entity.addRider(e.player));
                }
            });
        },
        onTick: e => {
            if (e.block.typeId != "mship:swivel_cannon") return;
            let dir = e.block.getState("minecraft:cardinal_direction");
            let a = { "west": -90, "north": 0, "east": 90, "south": 180 }[dir], hasCannon = false;
            e.block.dimension.getEntitiesAtBlockLocation(e.block).forEach(entity => {
                if (entity.hasFamily("cannon")) {
                    entity.setDynamicProperty("location", e.block.bottomCenter());
                    if (hasCannon) entity.remove(); // remove duplicates
                    else hasCannon = true;
                }
            });
            if (!hasCannon) e.block.dimension.runCommandAsync(`summon mship:swivel_cannon ${e.block.x + 0.5} ${e.block.y} ${e.block.z + 0.5} ${a} 0`);
        }
    });
});

world.afterEvents.dataDrivenEntityTrigger.subscribe(e => {
    try {
        let { entity, eventId } = e;
        if (!(entity && entity.isValid())) return;
        if (entity.typeId === "mship:swivel_cannon") {
            if (eventId == "mship:shoot") {
                const dist = 2.2, vel = 1.5, id = "mship:cannonball_projectile";
                const r = { x: -entity.getProperty("mship:pitch") * _PI_180_, y: -entity.getRotation().y * _PI_180_ };
                const dir = new Vector(Math.sin(r.y) * Math.cos(r.x), Math.sin(r.x), Math.cos(r.y) * Math.cos(r.x));
                const cannonball = entity.dimension.spawnEntity(id, entity.v3().above().add(dir.clone().mult(dist)));
                cannonball.getComponent("minecraft:projectile").shoot(dir.clone().mult(vel), { uncertainty: 0.1 });
                cannonball.addTag("swivel_cannon");
            }
        }
    } catch (err) { throw err; }
});

export default {};