import { BlockPermutation, system, Vector, world, _PI_180_ } from "./lib/Util";
import valid_blocks from "./valid_blocks";
import block_data from "./block_data";
import { updateFragmentsVisibility } from "./entities/fragments";

export default async function (player, block, data) {
    if (!data) return;
    const fragments = {}, dimension = player.dimension;
    let ship, fragment_collector, pos = block.below(data.y).bottomCenter(), lowest_buoy = 0;
    pos.x -= Math.cos((data.angle + 90) * _PI_180_);
    pos.z -= Math.sin((data.angle + 90) * _PI_180_);

    await dimension.runCommand(`summon mship:ship ${pos.x} ${pos.y} ${pos.z} ${data.angle} 0`);
    await dimension.runCommand(`summon mship:fragment_collector ${pos.x} ${pos.y} ${pos.z} ${data.angle} 0`);

    dimension.getEntities({ framilies: ["mship"], excludeFamilies: ["fragment"], excludeTags: ["activated"], location: pos, maxDistance: 1 }).some(e => {
        if (e?.typeId == 'mship:ship' && !ship) ship = e;
        else if (e?.typeId == 'mship:fragment_collector' && !fragment_collector) fragment_collector = e;
    });

    if (!ship || !fragment_collector) {
        if (!ship) sendMessage(player, "Failed to spawn ship".red());
        if (!fragment_collector) sendMessage(player, "Failed to spawn fragment controller".red());
        dimension.getEntities({ families: ["mship"], excludeFamilies: ["fragment"], excludeTags: ["activated"] }).some(e => e.remove());
        return;
    }

    fragment_collector.addTag(`captain=${player.id}`);
    fragment_collector.addTag(`ship=${ship.id}`);
    fragment_collector.getComponent("skin_id").value = valid_blocks.indexOf(block.typeId);

    spawnFragment("mship:fragment_helm").getComponent("minecraft:skin_id").value = valid_blocks.indexOf(block.typeId);
    dimension.spawnEntity("mship:sail_controller", block.bottomCenter()).setDynamicProperty("mship:sailToggle", false);
    const counter = dimension.spawnEntity("mship:particle", block.bottomCenter());
    const bitik = dimension.spawnEntity("mship:bitik", block.bottomCenter());
    bitik.addTag(`ship=${ship.id}`);
    
    ship.addTag(`captain=${player.id}`);
    ship.addTag(`ship=${ship.id}`);
    ship.getComponent('rideable').addRider(fragment_collector);
    ship.getComponent('rideable').addRider(player);
    ship.triggerEvent(`mship:y${data.y}`);
    ship.setDynamicProperty("mship:dataId", data.id);

    data.shipId = ship.id;
    data.entities = [];
    data.dblocks = [];
    data.y = data.lowestY;
    
    world.clearStructure(data.id);
    world.clearStructure(`${data.id}del`);
    let struc = world.createStructure(data.id, dimension, data.minPos, data.maxPos, { saveMode: "World", includeBlocks: true, includeEntities: false });
    let del = world.createStructure(`${data.id}del`, dimension, data.minPos, data.maxPos, { saveMode: "Memory", includeBlocks: true, includeEntities: false }), index = 0

    const cull = system.runJob(culltructure());
    function* culltructure() {
        try {
            for (let i = 0; i < struc.length; i++) {
                let s = struc[i], min = Vector.add(data.minPos, Vector.fromLongString(s.id.match(/\[(.*?)\]/)[1]).mult(64));
                for (let y = 0; y < s.size.y; y++) {
                    for (let z = 0; z < s.size.z; z++) {
                        for (let x = 0; x < s.size.x; x++) {
                            bitik.teleport(Vector.add(min, { x, y, z }));
                            let b = dimension.getBlock(Vector.add(min, { x, y, z }));
                            if (b && b.isValid()) {
                                let v = Vector.sub(b, block);
                                if (data.blocks[v.toLongString()] && data.valid[valid_blocks.indexOf(b.typeId)] == 1) {
                                    if (!block.hasTag("helm") || !ship.isValid() || !fragment_collector.isValid() || !player.isValid()) return failed();

                                    const bed = (b.typeId !== "minecraft:bed" || (b.typeId === "minecraft:bed" && b.getState("head_piece_bit")));
                                    const door = (!b.typeId.endsWith("_door") || (b.typeId.endsWith("_door") && !b.getState("upper_block_bit")));
                                    const bdata = block_data(b);
                                    if (b.typeId !== "minecraft:air" && bed && door) {
                                        if (!fragments[bdata.group]) fragments[bdata.group] = spawnFragment(`mship:blocky${bdata.group}`);
                                        bdata.fragmentId = fragments[bdata.group].id;
                                        bdata.fragmentSlot = fragments[bdata.group].count;

                                        if (b.getState("mship:fuel") > 0)
                                            setDynamicBlocks(fragments[bdata.group], "furnace_engine", { x, y, z }, v, s.id);
                                        else if (b.typeId.isOnList(['lit_furnace', 'lit_blast_furnace', 'lit_smoker', 'glowstone', 'lit_redstone_lamp', 'lantern', 'shroomlight', 'lit_pumpkin', 'froglight', 'torch', "end_rod"]))
                                            setDynamicBlocks(fragments[bdata.group], "light_source", { x, y, z }, v, s.id);

                                        let stack = (fragments[bdata.group].getDynamicProperty("mship:stack") || "") + createMolang(fragments[bdata.group].count, { _skin: bdata.skin, _data: bdata.data, px: v.x, py: v.y, pz: -v.z, rx: bdata.rot.x, ry: bdata.rot.y, rz: bdata.rot.z, v: 0 });
                                        fragments[bdata.group].setDynamicProperty("mship:stack", stack);

                                        fragments[bdata.group].count++;
                                        if (fragments[bdata.group].count >= 16) delete fragments[bdata.group];
                                    }
                                    if (b.hasTag("helm")) s.setBlockPermutation({ x, y, z }, s.getBlockPermutation({ x, y, z }).edit({ "mship:used": 0 }));
                                    else if (b.typeId == "minecraft:wet_sponge") s.setBlockPermutation({ x, y, z }, BlockPermutation.resolve("minecraft:sponge", null));
                                    bdata.position = v;
                                    bdata.structureId = s.id;
                                    bdata.structureOffset = { x, y, z };
                                    data.blocks[v.toLongString()] = bdata;

                                    checkPassengers(b.above());

                                    if (b.hasTag("buoy")) lowest_buoy = Math.min(lowest_buoy, v.y);
                                    counter.nameTag = `Compiling\n${Math.floor(100 * (++index / Object.keys(data.blocks).length))}%`;
                                    del[i].setBlockPermutation({ x, y, z }, BlockPermutation.resolve("minecraft:air", null));
                                } else {
                                    s.setBlockPermutation({ x, y, z }, null);
                                    if (data.blocks[v.toLongString()]) delete data.blocks[v.toLongString()];
                                }
                            }
                            yield;
                        }
                    }
                }
            }
            system.run(complete);
        } catch (err) { throw err; }
    }

    function checkPassengers(b) {
        let v = Vector.sub(b, block);
        dimension.getEntitiesAtBlockLocation(b).some(e => {
            if (e.hasFamily("fragment") || e.hasFamily("fragment_collector") || e.hasFamily("particle") || e.typeId === "mship:ship") return;
            if (e.typeId == "minecraft:player" || e.typeId == "minecraft:item" || e.hasTag("mship:isOnShip")) return;
            if (e.getRide() && e.getRide().typeId === "mship:seat") return;
            if (e.typeId === "mship:bitik") return;
            let pos = Vector.sub(e.v3(), block.bottomCenter()),
                d = {
                    id: e.id,
                    typeId: e.typeId,
                    distance: Math.sqrt(pos.x * pos.x + pos.z * pos.z),
                    angle: data.angle * _PI_180_ - Math.atan2(pos.z, pos.x),
                    baseAngle: data.angle,
                    offset: v.below(),
                    position: pos,
                    formerAngle: e.getRotation().y
                }
            if (e.hasFamily("cannon")) {
                if (e.typeId === "mship:cannon" && data.valid[valid_blocks.indexOf("mship:cannon_block")] == 0) return;
                if (e.typeId === "mship:swivel_cannon" && data.valid[valid_blocks.indexOf("mship:swivel_cannon")] == 0) return;
                if (e.typeId == "mship:cannon") d.formerAngle = { "west": -90, "north": 0, "east": 90, "south": 180 }[b.getState("minecraft:cardinal_direction")];
                d.distance = Math.sqrt(v.x * v.x + v.z * v.z);
                d.position = v;
            }
            e.addTag(`ship=${ship.id}`);
            e.addTag("mship:isActive");
            e.addTag("mship:isOnShip");
            if (e.getComponent("minecraft:mark_variant")) e.getComponent("minecraft:mark_variant").value = 1;
            data.entities.push(d);
        });
    }

    function setDynamicBlocks(fragment, type, offset, position, structureId) {
        let d = {
            type, fragmentId: fragment.id, fragmentSlot: fragment.count, structureId,
            distance: Math.sqrt(position.x * position.x + position.z * position.z),
            angle: data.angle * _PI_180_ - Math.atan2(position.z, position.x),
            position, offset
        }
        data.dblocks.push(d);
    }
    
    function spawnFragment(n) {
        const fragment = dimension.spawnEntity(n, block.bottomCenter());
        fragment_collector.addRider(fragment).then(_ => fragment.addTag('mship:isActive'));
        fragment.getComponent("minecraft:mark_variant").value = -data.angle;
        fragment.addTag(`ship=${ship.id}`);
        fragment.addEffect("invisibility", 99999, { showParticles: false });
        fragment.count = 0;
        return fragment;
    }

    function createMolang(index, list) {
        let str = "";
        for (let i in list) str += `v.b${index}${i}=${list[i]};`;
        return str;
    }

    function complete() {
        try {
            if (!block.hasTag("helm") || !ship.isValid() || !fragment_collector.isValid() || !player.isValid()) return failed();
            bitik.triggerEvent("mship:despawn");
            world.placeStructure(`${data.id}del`, dimension, data.minPos);
            world.clearStructure(`${data.id}del`);

            updateFragmentsVisibility(ship, data);

            if (lowest_buoy != 0) data.y = -lowest_buoy;
            ship.state = data.state;
            ship.triggerEvent(`mship:gravity_${data.state !== "default" ? "off" : "on"}`);
            ship.triggerEvent(`mship:${data.state === "diving" ? "submareal" : "watersheep"}`);
            ship.triggerEvent(`mship:y${data.y}`);
            ship.triggerEvent(`mship:control_on`);
            ship.runCommandAsync(`effect @e[tag="ship=${ship.id}"] clear`);
            ship.runCommandAsync(`tag @e[tag="ship=${ship.id}"] add mship:isActive`);
            ship.runCommandAsync(`event entity @e[type=mship:bitik,tag="ship=${ship.id}"] mship:despawn`);
            ship.getComponent("minecraft:mark_variant").value = data.y;
            ship.addTag("mship:isActive");
            ship.teleport(Vector.set(pos).above(0.174));
            world.clearProperty(data.id).setProperty(data.id, data);
            counter.nameTag = "Complete";
            counter.triggerEvent("mship:complete");
            sendMessage(player, "Complete");
        } catch (err) {
            failed();
            throw err;
        }
    }

    function failed() {
        sendMessage(player, "Failed".red());
        counter.nameTag = "Failed";
        bitik.triggerEvent("mship:despawn");
        counter.triggerEvent("mship:complete");
        system.clearJob(cull);
        world.clearStructure(data.id);
        world.clearStructure(`${data.id}del`);
        dimension.runCommandAsync(`event entity @e[ship=${shipId}] mship:despawn`);
        dimension.getEntities({ tags: [`ship=${shipId}`] }).forEach(e => e.remove());
    }
}

function sendMessage(player, msg) {
    if (player) player.sendMessage(`${msg}`);
}