import { _PI_180_, system, Vector, world, _180_PI_, getEntity } from "../lib/Util";
import valid_blocks from "../valid_blocks";

export function updateFragmentsVisibility(ship, d) {
    try {
        if (!ship.isValid()) return;
        const data = d || world.getParsedProperty(ship.getDynamicProperty("mship:dataId")), toUpdate = {};
        if (!data) return;
        for (let i in data.blocks) {
            let block = data.blocks[i];
            if (!toUpdate[block.fragmentId]) toUpdate[block.fragmentId] = {};
            toUpdate[block.fragmentId][`${block.fragmentSlot}v`] = 0;
            ["above", "below", "north", "south", "west", "east"].some(side => {
                let n = data.blocks[Vector.set(block.position)[side]().toLongString()];
                if (n?.solid != 2 || ((side === "above" || side === "below") && (n?.solid != 2 && n?.solid != 1))) {
                    toUpdate[block.fragmentId][`${block.fragmentSlot}v`] = 1;
                    return true;
                }
            });
        }

        for (let id in toUpdate) {
            let fragment = getEntity(id);
            if (fragment && fragment.isValid()) {
                let stack = fragment.getDynamicProperty("mship:stack");
                if (!stack) continue;
                for (let i in toUpdate[id]) {
                    stack = stack.replace(new RegExp(`(v.b${i}=)\\d+;`), `$1${toUpdate[id][i]};`);
                }
                fragment.playAnimation("animation.fragment.update", { stopExpression: `${stack}return !${Math.random()};` });
                fragment.setDynamicProperty("mship:stack", stack);
            }
        }
    } catch (err) { throw err; }
}

export function updateFragment(id, slot, list) {
    try {
        let fragment = getEntity(id);
        if (fragment && fragment.isValid()) {
            let stack = fragment.getDynamicProperty("mship:stack");
            if (!stack) return;
            for (let i in list) {
                stack = stack.replace(new RegExp(`(v.b${slot}${i}=)\\d+;`), `$1${list[i]};`);
            }
            fragment.playAnimation("animation.fragment.update", { stopExpression: `${stack}return !${Math.random()};` });
            fragment.setDynamicProperty("mship:stack", stack);
        }
    } catch (err) { throw err; }
}

world.afterEvents.dataDrivenEntityTrigger.subscribe(e => {
    try {
        let { entity, eventId } = e;
        if (!(entity && entity.isValid())) return;
        if (eventId == "mship:render") {
            let stack = entity.getDynamicProperty("mship:stack");
            if (stack) entity.playAnimation("animation.fragment.update", { stopExpression: `${stack}return !${Math.random()};` });
        }
    } catch (err) { throw err; }
});

world.beforeEvents.playerBreakBlock.subscribe(e => destroyBlock(e, e.block, e.player.dimension, e.player, "player"));
world.afterEvents.explosion.subscribe(e => e.getImpactedBlocks().forEach(b => destroyBlock(e, b, e.dimension, null, "explosion")));
function destroyBlock(e, block, dimension, player, n) {
    let blocks = [block], cancel = false;
    if (block.typeId.endsWith("_door")) {
        if (block.getState("upper_block_bit")) blocks.push(block.below());
        else blocks.push(block.above());
    }
    else if (block.typeId.endsWith(":bed")) {
        let other = [block.south(), block.west(), block.north(), block.east()][block.getState("direction") || 0];
        if (block.getState("head_piece_bit")) other = [block.north(), block.east(), block.south(), block.west()][block.getState("direction") || 0];
        blocks.push(other);
    }
    else if (block.typeId === "mship:cannon_block") blocks.push(block[block.getState("minecraft:cardinal_direction")]());

    blocks.some(b => {
        world.getPropertyIds().some(id => {
            const data = JSON.parse(world.getProperty(id) || null);
            if (!data) return;
            const v = Vector.sub(b, data.location);
            if (n === "player") {
                if (player) {
                    if (data.locked && player.id !== data.owner) {
                        player.sendMessage(`Locked`.red());
                        return (cancel = true);
                    }
                } else if (data.locked) return (cancel = true);
            }
            if (b.v3().equal(data.location)) {
                world.clearProperty(data.id);
                const owner = getEntity(data.owner);
                if (owner) owner.sendMessage(`${data.name} was destroyed`);
                return true;
            }
            if (data.blocks[v.toLongString()]) {
                data.cache.push(v.toLongString());
                delete data.blocks[v.toLongString()];

                data.minPos = new Vector(Math.min(data.minPos.x, b.x), Math.min(data.minPos.y, b.y), Math.min(data.minPos.z, b.z));
                data.maxPos = new Vector(Math.max(data.maxPos.x, b.x), Math.max(data.maxPos.y, b.y), Math.max(data.maxPos.z, b.z));

                system.run(_ => dimension.spawnEntity("mship:particle<mship:severe>", b.v3().add(0.5)));
                world.clearProperty(data.id);
                world.setProperty(data.id, data);
                return true;
            }
        });
    });
    if (typeof e.cancel === "boolean") return e.cancel = cancel;
}

world.afterEvents.playerPlaceBlock.subscribe(e => {
    let { block, player } = e;
    if (valid_blocks.includes(block.typeId)) {
        addBlock(player, block);
        if (e.block.typeId === "minecraft:bed") {
            let block = [e.block.south(), e.block.west(), e.block.north(), e.block.east()][e.block.getState("direction") || 0];
            addBlock(player, block);
        }
        else if (e.block.typeId.includes("_door")) addBlock(player, e.block.above());
        else if (e.block.typeId === "mship:cannon_block") addBlock(player, block[e.block.getState("minecraft:cardinal_direction")]());
    }
});

function addBlock(player, b) {
    world.getPropertyIds().some(id => {
        let data = JSON.parse(world.getProperty(id) || null);
        if (!data) return;
        let v = Vector.sub(b, data.location), isConnected = false;
        [b.above(), b.below(), b.north(), b.south(), b.west(), b.east()].some(neigbhor => {
            return isConnected = data.blocks[Vector.sub(neigbhor, data.location).toLongString()] != undefined;
        });
        if (isConnected) {
            if (data.locked && player.id !== data.owner) return true;
            if (data.valid[valid_blocks.indexOf(b.typeId)] != 1) return true;
            data.blocks[v.toLongString()] = { position: v };
            if (data.cache.includes(v.toLongString())) data.cache.splice(data.cache.indexOf(v.toLongString()), 1);

            data.minPos = new Vector(Math.min(data.minPos.x, b.x), Math.min(data.minPos.y, b.y), Math.min(data.minPos.z, b.z));
            data.maxPos = new Vector(Math.max(data.maxPos.x, b.x), Math.max(data.maxPos.y, b.y), Math.max(data.maxPos.z, b.z));

            player.dimension.spawnEntity("mship:particle<mship:merge>", b.v3().add(0.5));
            world.clearProperty(data.id);
            world.setProperty(data.id, data);
            return true;
        }
    });
}

export default { updateFragment, updateFragmentsVisibility };