import { system, _PI_180_, Vector, getEntity, angleDifference } from "../lib/Util";
 
export function updatePassengers(ship, data) {
    try {
        if (!(ship && ship.isValid())) return;
        if (!data) return;

        const shipAngle = ship.currentYaw * _PI_180_;
        const v = ship.v3().add(Math.cos((ship.currentYaw + 90) * _PI_180_), data.y, Math.sin((ship.currentYaw + 90) * _PI_180_));

        for (let d of data.entities) {
            let e = getEntity(d.id);
            if (!(e && e.isValid())) continue;
            const rotation = e.getRotation();
            const location = Vector.add(v, d.distance * Math.cos(shipAngle - d.angle), d.position.y, d.distance * Math.sin(shipAngle - d.angle));
            
            rotation.y = (ship.currentYaw - d.baseAngle) + d.formerAngle;

            if (e.typeId === "mship:swivel_cannon") updateSwivelCannon(e, ship, rotation, location);
            else e.teleport(location, { rotation });
        }
    } catch (err) { throw err; }
}

function updateSwivelCannon(e, ship, rotation, location) {
    if (e._yaw == undefined) e._yaw = rotation.y;
    if (e.formerYaw == undefined) e.formerYaw = 0;
    if (e.turnTick == undefined) e.turnTick = 0;

    if (ship.isTurning) {
        e.triggerEvent("mship:lock");
        e.turnTick = 0;
        delete e.riderYaw;
    }
    else {
        e.triggerEvent("mship:unlock");
        if (e.riderYaw == undefined) e.riderYaw = rotation.y;
        const player = e.getRider("minecraft:player");
        if (player) e.formerYaw = angleDifference(e.riderYaw, player.getRotation().y);
    }
    rotation.y += e.formerYaw;

    if (!ship.isMoving && e.turnTick > 0) {
        let r = e.getRotation().y * _PI_180_;
        e.applyImpulse(new Vector(0.0015 * Math.sin(r), 0, 0.0015 * Math.cos(r)));
        if (system.currentTick % 10 == 0) {
            e.teleport(location, { rotation });
            e.turnTick -= 1;
        }
    } else {
        if (ship.isMoving) e.turnTick = 0;
        e.teleport(location, { rotation });
    }

    if (e._yaw != e.getRotation().y) {
        e.turnTick = 2;
        e._yaw = e.getRotation().y;
    }
}

export default { updatePassengers };