import { _PI_180_, system, world, _180_PI_, clamp, Vector, angleDifference, ItemStack, getEntity, HALF_PI } from "../lib/Util";
import { VERSION, newVersionDetected, getCannons, getShips } from "../index";
import { clearWorldData, globalMenu } from "../ui/world";

function playerRidingCannon(player) {
    try {
        const ride = player.getRide();
        if (ride && ride.hasFamily("cannon")) {
            if (player.isJumping && !ride.getComponent("minecraft:is_ignited")) {
                if (ride.isUnderWater) return player.onScreenDisplay.setActionBar(`Can't fire Underwater`);
                if (player.getGameMode() === "survival") {
                    let list = player.findItem("mship:cannon_ball");
                    if (list.length > 0) {
                        player.setItem(list[0].item.setAmount(-1), list[0].slot);
                        ride.triggerEvent("mship:fire");
                    } else player.onScreenDisplay.setActionBar(`No Ammo`);
                } else ride.triggerEvent("mship:fire");
            }
            ride.setProperty("mship:pitch", clamp(player.getRotation().x, -45.0, 22.5) - 4);

            if (ride.typeId === "mship:swivel_cannon") {
                if (!ride.hasTag("mship:isOnShip")) {
                    if (Math.abs(angleDifference(ride.getRotation().y, player.getRotation().y)) > 0.25) {
                        let r = -player.getRotation().y * _PI_180_;
                        ride.applyImpulse(new Vector(0.0015 * Math.sin(r), 0, 0.0015 * Math.cos(r)));
                        if (system.currentTick % 10 == 0) ride.teleport(ride.getDynamicProperty('location'));
                    } else ride.teleport(ride.getDynamicProperty('location'));
                }
            }
        }
    } catch (err) {
        throw err;
    }
}

function playerSpawnPoint(player) {
    try {
        if (!player.getSpawnPoint()) return;
        let sp = Vector.set(player.getSpawnPoint()).toLongString();
        if (sp != player.getDynamicProperty("mship:spawnPoint")) {
            player.setDynamicProperty("mship:spawnPoint", sp);
            player.setSpawnPoint(player.getSpawnPoint());
            const block = player.dimension.getBlock(player.getSpawnPoint()).below();
            if (block) {
                world.getDynamicPropertyIds().some(a => {
                    if (!a.endsWith("_part0")) return;
                    const data = world.getParsedProperty(a.replace("_part0", ""));
                    if (!data) return;
                    if (!data.spawnPoints) return;
                    delete data.spawnPoints[player.id];
                    let v = Vector.sub(block, data.location);
                    if (!data.blocks[v.toLongString()]) return;
                    let p = v.clone();
                    p.distance = Math.sqrt(v.x * v.x + v.z * v.z);
                    p.angle = data.angle * _PI_180_ - Math.atan2(v.z, v.x);
                    data.spawnPoints[player.id] = p;
                    world.clearProperty(data.id).setProperty(data.id, data);
                    player.sendMessage(`Spawn Point set`);
                });
            }
        }
    } catch (err) {
        throw err;
    }
}

function playerCommands(player) {
    try {
        if (player.hasTag("cmd:worldSettings")) {
            player.removeTag("cmd:worldSettings");
            globalMenu(player);
        }
        if (player.hasTag("cmd:clearWorldData"))
            clearWorldData(player).removeTag("cmd:clearWorldData");
    } catch (err) {
        throw err;
    }
}

function playerHoldingCompass(player) {
    try {
        player.findItem("mship:ship_compass").forEach(c => {
            const n = Number(c.item.typeId.replace(/mship:ship_compass/g, "")) || 0;
            const dataId = c.item.getDynamicProperty("mship:dataId") || "";
            let targetLocation;
            for (let i in getShips()) {
                const ship = getShips()[i];
                if (!(ship && ship.isValid())) continue;
                if (ship.getDynamicProperty("mship:dataId") === dataId) targetLocation = ship.location;
            }
            if (!targetLocation) {
                const data = world.getParsedProperty(dataId);
                if (data) targetLocation = Vector.set(data.location);
            }
            if (targetLocation) {
                const pos = player.v3().sub(targetLocation.x + 0.5, 0, targetLocation.z + 0.5);
                const a = Math.atan2(pos.z, pos.x) * _180_PI_;
                let angDiff = angleDifference(a, player.getRotation().y - 90);
                if (angDiff < 0) angDiff = 360 + angDiff;
                const d = Math.round((angDiff / 360) * 16) % 16;

                if (d != n) {
                    let newItem = new ItemStack(`mship:ship_compass${d}`);
                    newItem.nameTag = c.item.nameTag;
                    newItem.setDynamicProperty("mship:dataId", dataId);
                    player.setItem(newItem, c.slot);
                }
            }
            else {
                let newItem = new ItemStack(`mship:broken_ship_compass`);
                newItem.nameTag = c.item.nameTag || "Ship Compass";
                player.setItem(newItem, c.slot);
            }
        });
    } catch (err) {
        throw err;
    }
}

function playerPointingBlock(player) {
    try {
        let str = `${player.getRotation().y}`;
        const block = player.getBlockFromViewDirection({ maxDistance: 10 }) || {};
        if (block.block != undefined && block.block.isValid()) {
            const b = block.block.above();
            str = `${b.typeId}\n` + 'Permutations:\n';
            for (let s in b.permutation.getAllStates()) {
                if (b.permutation.getState(s)) str += `${s} = ${b.permutation.getState(s)}\n`;
            }
            str += `Tags:\n`;
            b.getTags().forEach(tag => {
                str += `${tag}\n`;
            });
        }
        player.onScreenDisplay.setActionBar(`${str}`);
    } catch (err) {
        throw err;
    }
}

function test(player) {
    try {

        player.dimension.getEntities({ type: "minecraft:armor_stand" }).some(e => {
            let n = player.v3().sub(e.v3()).getYawPitch();
            n.yaw *= _180_PI_;

            let yaw = n.yaw - player.getRotation().y;
            yaw = (yaw < -180) ? (yaw + 360) : yaw;
            player.onScreenDisplay.setActionBar(`${yaw}`);

        });

    } catch (err) {
        throw err;
    }
}

export function updatePlayers() {
    let playerHoldingRemote = {};
    for (let player of world.getPlayers()) {
        playerHoldingCompass(player);
        playerCommands(player);
        playerRidingCannon(player);
        playerSpawnPoint(player);
        //playerPointingBlock(player);
        //test(player);
        
        try {
            if (player.getDynamicProperty("mship:nameTag") == undefined) player.setDynamicProperty("mship:nameTag", player.nameTag); // Store default nametag
            else if (player.nameTag != player.getDynamicProperty("mship:nameTag")) player.nameTag = player.getDynamicProperty("mship:nameTag"); // use stored nametag
            if (system.currentTick % 20 == 0) player.findItem("mship:cannon_block").forEach(i => player.setItem("mship:cannon", i.slot));

            if (player.hasTag("mship:first_spawn")) {
                player.removeTag("mship:first_spawn");
                if (newVersionDetected()) {
                    player.sendMessage(`New Version detected `.red() + `v[${VERSION}]`.green());
                    player.sendMessage(`reverting all ships and clearing world data is `.red() + `\n${"strongly recommended".strong()}`.red());
                    player.sendMessage(`/function revertAll`.red());
                    player.sendMessage(`/function clearWorldData`.red());
                }
            }
            
            const item = player.getItem();
            if (item?.typeId == "mship:cannon_remote") {
                const remoteId = item.getDynamicProperty("mship:remoteId");
                if (remoteId) {
                    if (playerHoldingRemote[remoteId] == undefined) playerHoldingRemote[remoteId] = [];
                    playerHoldingRemote[remoteId].push(player.nameTag);
                }
            }

            const ride = player.getRide();
            if (ride && ride.hasFamily("mship") && !ride.hasFamily("seat")) {
                player.playAnimation("animation.mship.on_helm", { "stopExpression": `return !q.is_riding;` });
                if (ride.typeId === "mship:ship") {
                    player.addTag("mship:restricMovements");
                    //player.inputPermissions.setPermissionCategory(11, false);
                    //player.inputPermissions.setPermissionCategory(12, false);
                }
            }
            if (!ride || !(ride && ride.hasFamily("ship"))) {
                if (player.hasTag("mship:hasCustomCamera")) {
                    player.runCommandAsync("camera @s clear");
                    player.removeTag("mship:hasCustomCamera");
                }
                if (player.hasTag("mship:restricMovements")) {
                    player.removeTag("mship:restricMovements");
                    //player.inputPermissions.setPermissionCategory(11, true);
                    //player.inputPermissions.setPermissionCategory(12, true);
                }
                player.findItem("mship:controller").forEach(item => player.setItem(null, item.slot));
            }
        } catch (err) { throw err; }
    }

    for (let i in getCannons()) {
        const cannon = getCannons()[i];
        if (!(cannon && cannon.isValid())) continue;
        const remoteId = cannon.getDataFromTag("remoteId");
        const players = playerHoldingRemote[remoteId] || [];
        cannon.playAnimation("animation.mship.cannon.hide_remote", { stopExpression: "q.is_alive == 0" });
        cannon.playAnimation("animation.mship.cannon.show_remote", { stopExpression: "q.is_alive == 0", players });
    }
}

world.afterEvents.playerSpawn.subscribe(e => {
    e.initialSpawn = false;
    world.getDynamicPropertyIds().some(a => {
        if (!a.endsWith("_part0")) return;
        const data = JSON.parse(world.getProperty(a.replace("_part0", "")) || null);
        if (!data || !data?.spawnPoints) return;
        if (data.spawnPoints[e.player.id]) {
            const d = data.spawnPoints[e.player.id];
            let base = Vector.add(data.location, 0.5, 0, 0.5), baseAngle = data.angle * _PI_180_;
            let ship = getEntity(data.shipId);
            if (ship && ship.isValid()) {
                baseAngle = ship.currentYaw * _PI_180_;
                base = Vector.add(ship.v3(), Math.cos(baseAngle + HALF_PI), 0, Math.sin(baseAngle + HALF_PI));
            }
            const v = base.add(d.distance * Math.cos(baseAngle - d.angle), d.y + data.y, d.distance * Math.sin(baseAngle - d.angle));
            e.player.teleport(v);
            v.dimension = e.player.dimension;
            e.player.setSpawnPoint(v);
            e.player.setDynamicProperty("mship:spawnPoint", v.toLongString());
        }
    });
});

world.afterEvents.playerInteractWithEntity.subscribe(e => {
    if (e.target.typeId != "mship:sail_controller") return;
    if (e.itemStack) return;
    let ship = getEntity(e.target.getDataFromTag("ship"));
    if (!(ship && ship.isValid())) return;
    let captain = getEntity(ship.getDataFromTag("captain"));
    if (!(captain && captain.isValid())) return;
    if (e.player.id !== captain.id) return;
    e.target.setDynamicProperty("mship:sailToggle", !e.target.getDynamicProperty("mship:sailToggle"));
    ship.sailControl = e.target.getDynamicProperty("mship:sailToggle");
    ship.triggerEvent(`mship:control_${ship.sailControl ? "off" : "on"}`);
    e.target.nameTag = `${e.target.getDynamicProperty("mship:sailToggle") ? "LOCKED" : ""}`;
});

export default { updatePlayers };