import { Vector, world, _PI_180_ } from "../lib/Util";
import { getProjectiles } from "../index";
import { explode } from "../lib/Explosion";

world.afterEvents.projectileHitBlock.subscribe(e => {
    if (e.projectile.typeId === "mship:cannonball_projectile") e.projectile.triggerEvent("mship:explode");
});

world.afterEvents.projectileHitEntity.subscribe(e => {
    if (e.getEntityHit().entity.hasFamily("mship")) return;
    if (e.projectile.typeId === "mship:cannonball_projectile") e.projectile.triggerEvent("mship:explode");
});

export function projectileUpdate(ship, data) {
    try {
        if (Object.keys(getProjectiles()).length == 0) return;
        const angle = (data.angle - ship.currentYaw) * _PI_180_;
        const shipPos = ship.v3().add(Math.cos((ship.currentYaw + 90) * _PI_180_), data.y + 0.5, Math.sin((ship.currentYaw + 90) * _PI_180_));

        for (let p in getProjectiles()) {
            const proj = getProjectiles()[p];
            if (!(proj && proj.isValid())) continue;
            if (proj?.deflected) continue;
            if (proj.prev == undefined) proj.prev = proj.v3();

            const yp = Vector.sub(proj.v3(), proj.prev).getYawPitch();
            const n = translate(proj.prev, shipPos, angle);
            meep: for (let i = 0; i < 2; i++) {
                let v = Vector.add(n, Vector.applyAngle(yp.yaw, yp.pitch, i));
                if (ship.hashMap.isInside(v)) {
                    if (ship.hashMap.intersectsSphere(v).length > 0) {
                        if (proj.typeId === "mship:cannonball_projectile") {
                            proj.triggerEvent(`mship:explode`);
                            explode(ship, Vector.sub(v, Vector.applyAngle(yp.yaw, yp.pitch)), proj.hasTag("swivel_cannon") ? 2 : 3, data);
                        } else {
                            proj.teleport(Vector.add(proj.prev, Vector.applyAngle(yp.yaw, yp.pitch, i)), { keepVelocity: false });
                            proj.deflected = true;
                        }
                        break meep;
                    }
                }
            }
            proj.prev = proj.v3().clone();
        }
        /**/
    } catch (er) {
        throw er;
    }
}

function translate(p, origin, angle) {
    const offset = Vector.sub(p, origin);
    const yawDiff = Math.atan2(offset.z, offset.x) + angle;
    const mag = Math.sqrt(offset.x * offset.x + offset.z * offset.z);
    return new Vector(Math.cos(yawDiff) * mag, offset.y, Math.sin(yawDiff) * mag);
}

export default { projectileUpdate };