import { world, system, _180_PI_, angleDifference, Vector, _PI_180_, HALF_PI, getEntity } from "../lib/Util";
import { projectileUpdate } from "./projectiles";
import { SpatialHash } from "../lib/SpatialHash";
import { updatePassengers } from "./passengers";
import { updateFragment } from "./fragments";
import { mainMenu } from "../ui/main";
import { getShips } from "../index";
import { Box } from "../lib/Box";
import revert from "../revert";

function checkNearestPlayer(ship, data) {
    try {
        let tmp = {}, update = false, toUpdate = {};
        //const angle = ship.currentYaw * _PI_180_;
        //const shipPos = ship.v3().add(Math.cos((ship.currentYaw + 90) * _PI_180_), 0, Math.sin((ship.currentYaw + 90) * _PI_180_));
        for (let player of ship.dimension.getPlayers({ location: ship.location, maxDistance: 48 })) {
            if (!ship.nearestPlayers[player.id]) {
                ship.nearestPlayers[player.id] = player;
                update = true;
            }
            tmp[player.id] = 1;
        }
        for (let i in ship.nearestPlayers) {
            if (!tmp[i]) delete ship.nearestPlayers[i];
            //else {
            //    const player = ship.nearestPlayers[i];
            //    for (let b in data.blocks) {
            //        let block = data.blocks[b];

            //        const bAngle = data.angle * _PI_180_ - Math.atan2(block.position.z, block.position.x);
            //        const bDist = Math.sqrt(block.position.x * block.position.x + block.position.z * block.position.z);
            //        const v = Vector.add(shipPos, bDist * Math.cos(angle - bAngle), block.position.y + data.y, bDist * Math.sin(angle - bAngle));

            //        const yaw = angleDifference(player.v3().sub(v).getYawPitch().yaw * _180_PI_, player.getRotation().y);
            //        if (!toUpdate[block.fragmentId]) toUpdate[block.fragmentId] = {};
            //        toUpdate[block.fragmentId][`${block.fragmentSlot}v`] = (yaw > -130 && yaw < 130) ? 0 : 1;
            //    }
            //}
        }
        //for (let id in toUpdate) {
        //    let fragment = getEntity(id);
        //    if (fragment && fragment.isValid()) {
        //        let stack = fragment.getDynamicProperty("mship:stack");
        //        if (!stack) continue;
        //        for (let i in toUpdate[id]) {
        //            stack = stack.replace(new RegExp(`(v.b${i}=)\\d+;`), `$1${toUpdate[id][i]};`);
        //        }
        //        fragment.playAnimation("animation.fragment.update", { stopExpression: `${stack}return !${Math.random()};` });
        //        fragment.setDynamicProperty("mship:stack", stack);
        //    }
        //}
        if (update) ship.dimension.runCommandAsync(`event entity @e[family=fragment,tag="ship=${ship.id}"] mship:render`);
    }
    catch (er) { throw er; }
}

function setupHashMap(ship, data) {
    if (ship.hashMap) return;
    ship.hashMap = new SpatialHash(4);

    for (let i in data.blocks) {
        let box = new Box(Vector.add(data.blocks[i].position, 0.5)).setId(i).setData(data.blocks[i]);
        ship.hashMap.insert(box.setData(data.blocks[i]));
    }
}

function control(ship) {
    try {
        const vel = ship.velocityV3();
        const mag = Math.sqrt(Math.pow(vel.x, 2) + Math.pow(vel.z, 2));
        const heading = -Math.atan2(vel.x, vel.z) * _180_PI_;
        if (mag >= 0.005) ship.headingError = angleDifference(heading, ship.currentYaw);

        let speed = (world.getDynamicProperty("mship:land_speed") / 1e3) + ship.bonusSpeed;
        let waterSpeed = (world.getDynamicProperty("mship:water_speed") / 1e3) + ship.bonusSpeed;

        if (!ship.sailControl) {
            ship.velocity = ship.velocityV3();
            if (mag >= 0.05) ship.currentYaw = ship.getRotation().y;
            else ship.setRotation({ x: 0, y: ship.currentYaw });
        } else {
            let sp = 0.2;
            if (ship.state == "flying") sp = 0.1;
            ship.applyImpulse(Vector.mult(ship.velocity, sp));
        }

        if (ship.headingError <= -100 || ship.headingError >= 100) {
            speed *= 2;
            waterSpeed *= 2;
        }
        else if (ship.headingError < -32 || ship.headingError > 32) {
            speed = 0.01;
            waterSpeed = 0.01;
        }

        ship.isMoving = (mag >= 0.05) || Math.abs(ship._location.y - ship.location.y) >= 0.01;

        flightDiveMechanics(ship, speed, waterSpeed);

        const collector = ship.getRider("mship:fragment_collector");
        if (collector) collector.setRotation({ x: 0, y: ship.currentYaw });
    } catch (er) {
        throw er;
    }
}

function flightDiveMechanics(ship, speed, waterSpeed) {
    try {
        if (ship.hasFamily("sinkingsheep")) {
            if (ship.state == "flying") {
                ship.addEffect("slow_falling", 1e3, { amplifier: 3, showParticles: false });
                if (ship.isOnGround || ship.isInWater) ship.state = "default";
            }
            return;
        }

        const captain = ship.getRider("minecraft:player");
        if (captain && captain.isValid()) {

            if (captain.controllerSetting === "down") {
                captain.controllerSetting = "";
                if (ship.isInWater) {
                    if (ship.state != "diving") {
                        ship.triggerEvent("mship:gravity_off");
                        ship.triggerEvent("mship:submareal");
                    }
                    ship.state = "diving";
                }
            }
            else if (captain.controllerSetting === "up") {
                captain.controllerSetting = "";
                if (ship.state != "flying") {
                    ship.triggerEvent("mship:gravity_on");
                    ship.triggerEvent("mship:watersheep");
                }
                ship.state = "flying";
            }

            let item = captain.getItem(), text = "";
            if (item?.typeId === "mship:controller") {

                if (ship.state === "flying") {
                    ship.levitation = 3 + Math.max(Math.floor(-captain.getRotation().x / 8), 0);
                    if (ship.levitation - 3 > 0) {
                        ship.triggerEvent("mship:gravity_off");
                        text = "[Ascending] " + (ship.levitation - 3);
                        if (ship.isInWater) ship.applyImpulse({ x: 0, y: 0.05, z: 0 });
                    } else if (captain.getRotation().x > 32) {
                        text = `[Descending]`;
                        if (ship.isOnGround || ship.isInWater) {
                            ship.triggerEvent("mship:gravity_on");
                            ship.state = "default";
                        }
                        ship.levitation = 0;
                    } else ship.levitation = 3;
                }
                else if (ship.state === "diving") {
                    if (captain.getRotation().x < -24) {
                        ship.levitation = 0.01 * Math.max(Math.floor(-(captain.getRotation().x + 24) / 8), 0);
                        text = "[Ascending] " + Math.floor(ship.levitation * 100);
                    }
                    else if (captain.getRotation().x > 32) {
                        ship.levitation = -0.01 * Math.max(Math.floor((captain.getRotation().x - 32) / 8), 0);
                        text = "[Descending] " + Math.floor(-ship.levitation * 100);
                    }
                    else ship.levitation = 0;
                    if (!(ship.isOnGround || ship.isInWater)) {
                        ship.triggerEvent("mship:watersheep");
                        ship.triggerEvent("mship:gravity_on");
                        ship.state = "default";
                    }
                }
            } else ship.levitation = ship.state === "flying" ? 3 : 0

            if (ship.state !== "default") {
                captain.onScreenDisplay.setActionBar(`Altitude: ${~~ship.v3().y}\n${text}`);
                if (ship.state === "flying") speed += 0.5;
            }
            if (ship.state != "default" && captain.findItem("mship:controller").length == 0) captain.addItem("mship:controller");

        } else ship.levitation = ship.state === "flying" ? 3 : 0

        if (ship.state === "flying") ship.addEffect("levitation", 5, { amplifier: ship.levitation, showParticles: false });
        else if (ship.state === "diving") ship.applyImpulse(new Vector(0, ship.levitation, 0));

        ship.getComponent('minecraft:underwater_movement').setCurrentValue(waterSpeed / 2);
        ship.getComponent('minecraft:movement').setCurrentValue(speed);
    } catch (er) {
        throw er;
    }
}

function playerCamera(ship, data) {
    try {
        const captain = ship.getRider("minecraft:player");
        if (captain) {
            if (captain.isJumping) mainMenu(captain, ship);
            if (data.camera) {
                if (data.cameraDistance <= 5) captain.runCommandAsync("camera @s set minecraft:third_person");
                else captain.runCommand(`execute positioned ~~2~ run camera @s set minecraft:free ease 0.1 linear pos ^^^-${data.cameraDistance} rot ~~`);
                captain.addTag("mship:hasCustomCamera");
            } else if (captain.hasTag("mship:hasCustomCamera")) {
                captain.runCommandAsync("camera @s clear");
                captain.removeTag("mship:hasCustomCamera");
            }
        }
    } catch (er) {
        throw er;
    }
}

function updateDynamicBlocks(ship, data) {
    try {
        for (let d of data.dblocks) {
            let light = false;
            const shipAngle = ship.currentYaw * _PI_180_;
            const v = ship.v3().add(Math.cos(shipAngle + HALF_PI), 0, Math.sin(shipAngle + HALF_PI));
            v.add(d.distance * Math.cos(shipAngle - d.angle), d.position.y + data.y, d.distance * Math.sin(shipAngle - d.angle));
            const block = ship.dimension.getBlock(v);
            if (!(block && block.isValid())) continue;
            if (d.type === "light_source") light = true;
            else if (d.type === "furnace_engine") {
                let struc = world.structureManager.get(d.structureId);
                if (struc) {
                    let p = struc.getBlockPermutation(d.offset);
                    let s0 = p.getState("mship:seconds_0") || 0;
                    let s1 = p.getState("mship:seconds_1") || 0;
                    let fuel = p.getState("mship:fuel") || 0;
                    if (fuel > 0) {
                        if (system.currentTick % 20 == 0) {
                            fuel = Math.max(fuel - (~~((s0 + 1) / 10) * ~~((s1 + 1) / 8)), 0);
                            s1 = (s1 + ~~((s0 + 1) / 10)) % 8;
                            s0 = (s0 + 1) % 10;
                            updateFragment(d.fragmentId, d.fragmentSlot, { "_skin": 12 * 20 + fuel });
                            struc.setBlockPermutation(d.offset, p.edit({ "mship:seconds_0": s0, "mship:seconds_1": s1, "mship:fuel": fuel }));
                        }

                        let particle = block.above().hasTag("water") ? "mship:bubble_particle" : "minecraft:campfire_tall_smoke_particle";
                        if (particle === "mship:bubble_particle" && system.currentTick % 10 == 0) ship.dimension.spawnParticle(particle, v.above(1.2));
                        else if (particle === "minecraft:campfire_tall_smoke_particle" && system.currentTick % 5 == 0) ship.dimension.spawnParticle(particle, v.above(1.2));
                        light = true;
                        ship.bonusSpeed = 0.08;
                    }
                }
            }

            if (light && (block.typeId === "minecraft:air" || block.typeId === "mship:ilaw")) block.editState({ "mship:timer": 1 }, "mship:ilaw");
        }
    } catch (err) { throw err; }
}

export function updateShips() {
    for (let s in getShips()) {
        try {
            const ship = getShips()[s];
            if (!(ship && ship.isValid())) continue;
            if (!ship.hasTag("mship:isActive")) continue;
            if (ship.currentYaw == undefined) ship.currentYaw = ship.getRotation().y;
            if (ship._currentYaw == undefined) ship._currentYaw = ship.getRotation().y;
            if (ship._location == undefined) ship._location = ship.location;
            if (ship.bonusSpeed == undefined) ship.bonusSpeed = 0;
            if (ship.isTurning == undefined) ship.isTurning = false;
            if (ship.isMoving == undefined) ship.isMoving = false;
            if (ship.levitation == undefined) ship.levitation = 3;
            if (ship.nearestPlayers == undefined) ship.nearestPlayers = [];
            if (ship.velocity == undefined) ship.velocity = ship.velocityV3();
            if (ship.headingError == undefined) ship.headingError = 0;
            const data = world.getParsedProperty(ship.getDynamicProperty("mship:dataId"));
            if (!data) continue;

            if (ship.state == undefined) ship.state = data.state;

            setupHashMap(ship, data);
            updateDynamicBlocks(ship, data);
            control(ship);
            playerCamera(ship, data);
            updatePassengers(ship, data);
            
            if (system.currentTick % 20 == 0) {
                checkNearestPlayer(ship, data);
            }

            ship.isTurning = (Math.abs(ship._currentYaw - ship.currentYaw) >= 0.2);
            ship._currentYaw = ship.currentYaw;
            ship._location = ship.location;

            if (data.state != ship.state) {
                data.state = ship.state;
                world.setProperty(data.id, JSON.stringify(data));
            }

            ship.bonusSpeed = 0;
            projectileUpdate(ship, data);
        } catch (err) { throw err; }
    }
}

world.afterEvents.dataDrivenEntityTrigger.subscribe(e => {
    try {
        let { entity, eventId } = e;
        if (!(entity && entity.isValid())) return;
        if (entity.typeId === "mship:ship" && eventId == "mship:revert") revert(entity);
    } catch (err) { throw err; }
});

export default { updateShips }