import { _PI_180_, system, world, _180_PI_, OVERWORLD, getEntity } from "./lib/Util";

import { updateShips } from "./entities/ships";
import { updatePlayers } from "./entities/players";

import valid_blocks from "./valid_blocks";

import { } from "./blocks/helm";
import { } from "./blocks/furnace_engine";
import { } from "./blocks/seats";
import { } from "./blocks/light_block";
import { } from "./blocks/cannon_block";
import { } from "./blocks/swivel_cannon";
import { } from "./items/cannon_remote";
import { } from "./items/ship_compass";
import { } from "./items/controller";
export const VERSION = "0.70.0";

let NEW_VERSION_DETECTED = false;

const SHIPS = {};
const CANNONS = {};
const PROJECTILES = {};

const MIN_BLOCKS = 12;
const MAX_BLOCKS = 5124;

function checkEntity(e, opt) {
    const id = `${e.typeId}`;
    if (opt === "add") {
        if (id === "mship:ship") SHIPS[e.id] = e;
        else if (id === "mship:cannon" || id === "mship:swivel_cannon") CANNONS[e.id] = e;
        else if (e.hasComponent("minecraft:projectile")) PROJECTILES[e.id] = e;
    } else if (opt === "remove") {
        delete SHIPS[e];
        delete CANNONS[e];
        delete PROJECTILES[e];
    }
}

function startup() {
    OVERWORLD.getEntities().forEach(e => checkEntity(e, "add"));
    system.runInterval(_ => {
        checkData();
        updatePlayers();
        updateShips();
    }, 0);
}

function checkData() {
    if (world.getDynamicProperty("mship:version") !== VERSION) {
        world.setDynamicProperty("mship:valid_blocks", null);
        world.setDynamicProperty("mship:min_blocks", null);
        world.setDynamicProperty("mship:max_blocks", null);
        world.setDynamicProperty("mship:land_speed", null);
        world.setDynamicProperty("mship:water_speed", null);
        world.setDynamicProperty("mship:controller", null);
        world.setDynamicProperty("mship:version", null);
        console.warn(`New Version Found`);
        console.warn(`v${VERSION} data updated`);

        NEW_VERSION_DETECTED = true;
    }
    if (world.getDynamicProperty("mship:valid_blocks") == undefined) world.setDynamicProperty("mship:valid_blocks", "1".repeat(valid_blocks.length));
    if (world.getDynamicProperty("mship:min_blocks") == undefined) world.setDynamicProperty("mship:min_blocks", MIN_BLOCKS);
    if (world.getDynamicProperty("mship:max_blocks") == undefined) world.setDynamicProperty("mship:max_blocks", MAX_BLOCKS);
    if (world.getDynamicProperty("mship:land_speed") == undefined) world.setDynamicProperty("mship:land_speed", 1e2);
    if (world.getDynamicProperty("mship:water_speed") == undefined) world.setDynamicProperty("mship:water_speed", 1e2);
    if (world.getDynamicProperty("mship:controller") == undefined) world.setDynamicProperty("mship:controller", false);
    if (world.getDynamicProperty("mship:version") == undefined) world.setDynamicProperty("mship:version", VERSION);
}

world.afterEvents.worldInitialize.subscribe(startup);
world.afterEvents.entityLoad.subscribe(e => checkEntity(e.entity, 'add'));
world.afterEvents.entitySpawn.subscribe(e => checkEntity(e.entity, 'add'));
world.afterEvents.entityDie.subscribe(e => checkEntity(e.deadEntity.id, 'remove'));
world.beforeEvents.entityRemove.subscribe(e => checkEntity(e.removedEntity.id, 'remove'));

export function getShips() { return SHIPS; }
export function getCannons() { return CANNONS; }
export function getProjectiles() { return PROJECTILES; }
export function newVersionDetected() { return NEW_VERSION_DETECTED == true; };

export default { getShips, getCannons, getProjectiles, VERSION, newVersionDetected };