function pointInsideBox(point, box) {
    // Transform point into box's local coordinate system
    let localPoint = rotatePointAroundPoint(point, box.center, -box.yaw, -box.pitch);
    // Check if the transformed point lies within the box's extents
    return (
        Math.abs(localPoint.x - box.center.x) <= box.width / 2 &&
        Math.abs(localPoint.y - box.center.y) <= box.height / 2 &&
        Math.abs(localPoint.z - box.center.z) <= box.length / 2
    );
}

// Function to rotate a point around another point in 3D space
function rotatePointAroundPoint(point, center, yaw, pitch) {
    // Translate point to origin
    let translatedPoint = {
        x: point.x - center.x,
        y: point.y - center.y,
        z: point.z - center.z
    };

    // Rotate around y-axis (yaw)
    let newX = translatedPoint.x * Math.cos(yaw) - translatedPoint.z * Math.sin(yaw);
    let newZ = translatedPoint.x * Math.sin(yaw) + translatedPoint.z * Math.cos(yaw);

    translatedPoint.x = newX;
    translatedPoint.z = newZ;

    // Rotate around x-axis (pitch)
    let newY = translatedPoint.y * Math.cos(pitch) - translatedPoint.z * Math.sin(pitch);
    newZ = translatedPoint.y * Math.sin(pitch) + translatedPoint.z * Math.cos(pitch);

    translatedPoint.y = newY;
    translatedPoint.z = newZ;

    // Translate back to original position
    return {
        x: translatedPoint.x + center.x,
        y: translatedPoint.y + center.y,
        z: translatedPoint.z + center.z
    };
}

// Example usage
let box = {
    center: { x: 0, y: 0, z: 0 },
    width: 5,
    height: 3,
    length: 4,
    yaw: Math.PI / 4, // 45 degrees in radians
    pitch: Math.PI / 6 // 30 degrees in radians
};

let point = { x: 1, y: 1, z: 1 };

console.log(pointInsideBox(point, box)); // Output: true if point is inside the box
