import { clamp, getEntity, system, Vector, world, _PI_180_ } from "./Util";

const __dirs__ = createDirections();

export function explode(ship, position, power, d) {
    let rays = [];
    let expiredRaysCount = 0, cycleLimit = 20;
    let data = d || world.getParsedProperty(ship.getDynamicProperty("mship:dataId"));
    let helmBroke = false;

    function loop() {
        let toUpdate = {};
        let destroyed = {};

        if (!ship.isValid()) return;
        __dirs__.forEach((d, i) => {
            if (!rays[i]) rays[i] = new Ray(position, power * (0.7 + Math.random() * 0.6), d);
            if (rays[i].intensity <= 0) {
                expiredRaysCount += 1;
                return;
            }
            let id = rays[i].attenuate(data.blocks);
            if (id) {
                let block = data.blocks[id];
                if (!toUpdate[block.fragmentId]) toUpdate[block.fragmentId] = {};
                toUpdate[block.fragmentId][`${block.fragmentSlot}_skin`] = -1;

                ["above", "below", "north", "south", "west", "east"].some(side => {
                    let n = data.blocks[Vector.set(block.position)[side]().toLongString()];
                    if (n) {
                        if (!toUpdate[n.fragmentId]) toUpdate[n.fragmentId] = {};
                        toUpdate[n.fragmentId][`${n.fragmentSlot}v`] = 1;
                    }
                });

                if (id === "0,0,0" || block.typeId.includes("_helm")) {
                    if (!helmBroke) ship.runCommandAsync(`event entity @e[type=mship:fragment_helm,tag="ship=${ship.id}"] mship:despawn`);
                    helmBroke = true;
                }
                destroyed[id] = 1;
            }
        });

        for (let id in toUpdate) {
            let fragment = getEntity(id);
            if (fragment && fragment.isValid()) {
                let stack = fragment.getDynamicProperty("mship:stack");
                if (!stack) continue;
                for (let i in toUpdate[id]) {
                    stack = stack.replace(new RegExp(`(v.b${i}=)\\d+;`), `$1${toUpdate[id][i]};`);
                }
                fragment.playAnimation("animation.fragment.update", { stopExpression: `${stack}return !${Math.random()};` });
                fragment.setDynamicProperty("mship:stack", stack);
            }
        }

        for (let id in destroyed) {
            let block = data.blocks[id];
            let s = world.structureManager.get(block.structureId);
            if (s) s.setBlockPermutation(block.structureOffset, null);

            for (let i = data.entities.length - 1; i >= 0; i--) {
                let d = data.entities[i];
                if (Vector.set(d.offset).equal(block.position)) data.entities.splice(i, 1);
            }

            for (let i = data.dblocks.length - 1; i >= 0; i--) {
                let d = data.dblocks[i];
                if (Vector.set(d.position).equal(block.position)) data.dblocks.splice(i, 1);
            }

            if (!data.cache.includes(id)) data.cache.push(id);
            ship.hashMap.remove(id);
            delete data.blocks[id];
        }

        cycleLimit -= 1;
        if (expiredRaysCount == __dirs__.length || cycleLimit <= 0) {
            const length = Object.keys(data.blocks).length;
            const totalBlocks = Math.max(length + data.cache.length, 0);
            const integrity = clamp(~~(100 * (length / totalBlocks)), 0, 100);
            if (integrity < 40 || helmBroke) {
                ship.triggerEvent("mship:sink");
                const owner = getEntity(data.owner);
                if (owner) owner.sendMessage(`${data.name} was destroyed`);
            }
            world.clearProperty(data.id);
            world.setProperty(data.id, JSON.stringify(data));

            return;
        }
        system.run(loop);
    }
    loop();
}


function Ray(position, intensity, direction) {
    this.position = position.clone();
    this.intensity = intensity;
    this.direction = direction.clone();

    this.attenuate = function (blocks) {
        let result, id = this.position.floor().toLongString();
        if (blocks[id]) {
            this.intensity -= (blocks[id].resistance + 0.3) * 0.3;
            if (this.intensity > 0) result = id;
        }
        this.position.add(this.direction);
        this.intensity -= 0.225;
        return result;
    }
}  

function createDirections(steps = 10) {
    const dirs = [];
    for (let yawAngle = 0; yawAngle < 360; yawAngle += steps) {
        for (let pitchAngle = 0; pitchAngle < 360; pitchAngle += steps) {
            dirs.push(Vector.applyAngle(yawAngle * _PI_180_, pitchAngle * _PI_180_, 0.3));
        }
    }
    return dirs;
}

export default { explode };