

export function Quaternion(x, y, z, w) {
    this.x = x;
    this.y = y;
    this.z = z;
    this.w = w;
}
Quaternion.set = function (quaternion) {
    if (typeof quaternion != "object") return new Quaternion(0, 0, 0, 1);
    return new Quaternion(quaternion.x, quaternion.y, quaternion.z, quaternion.w);
}
Quaternion.prototype.clone = function () {
    return new Quaternion(this.x, this.y, this.z, this.w);
}
Quaternion.prototype.conjugate = function () {
    return new Quaternion(-this.x, -this.y, -this.z, this.w);
}
Quaternion.prototype.multiply = function (q) {
    const { x: a, y: b, z: c, w: d } = this;
    const { x: e, y: f, z: g, w: h } = q;

    const x = a * h + b * g - c * f + d * e;
    const y = -a * g + b * h + c * e + d * f;
    const z = a * f - b * e + c * h + d * g;
    const w = -a * e - b * f - c * g + d * h;

    return new Quaternion(x, y, z, w);
}
Quaternion.fromEuler = function (yaw, pitch, roll) {
    const cy = Math.cos(pitch * 0.5);
    const sy = Math.sin(pitch * 0.5);
    const cr = Math.cos(roll * 0.5);
    const sr = Math.sin(roll * 0.5);
    const cp = Math.cos(yaw * 0.5);
    const sp = Math.sin(yaw * 0.5);

    const qw = cy * cr * cp + sy * sr * sp;
    const qx = cy * sr * cp - sy * cr * sp;
    const qy = cy * cr * sp + sy * sr * cp;
    const qz = sy * cr * cp - cy * sr * sp;

    return new Quaternion(qx, qy, qz, qw);
}

export default { Quaternion };