import * as mc from '@minecraft/server';
import * as ui from '@minecraft/server-ui';
import String from './String'

export const world = mc.world;
export const system = mc.system;
export const BlockPermutation = mc.BlockPermutation;
export const ItemStack = mc.ItemStack;
export const Block = mc.Block;
export const BlockTypes = mc.BlockTypes;
export const Entity = mc.Entity;
export const Player = mc.Player;

export const ActionFormData = ui.ActionFormData;
export const ModalFormData = ui.ModalFormData;
export const uiManager = ui.uiManager;

export const _PI_180_ = Math.PI / 180;
export const _180_PI_ = 180 / Math.PI;
export const HALF_PI = Math.PI / 2;

const ALPHABET = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', NUMBERS = '0123456789', SPECIAL = '!@#$%^&*()_-+=<>?';

world.globalText = "";

Array.prototype.anyEqual = function () { return this.filter((item, index) => this.indexOf(item) !== index).length != 0; }
Array.prototype.allEqual = function () { return this.every(val => val === this[0]); }
Array.prototype.equals = function (array) {
    if (!array) return false;
    if (array === this) return true;
    if (this.length != array.length) return false;
    for (let i = 0; i < this.length; i++) {
        if (this[i] instanceof Array && array[i] instanceof Array) return this[i].equals(array[i]);
        if (this[i] !== array[i]) return false;
    }
    return true;
}
Math.randomInteger = function (min, max) { return min + ~~(Math.random() * (max - min + 1)); }

Block.prototype.toString = function () { return JSON.stringify({ ... this.location }); }
Block.prototype.v3 = function () { return Vector.set(this.location); }
Block.prototype.getState = function (state) { return this.permutation.getState(state); }
Block.prototype.getAllStates = function () {
    let states = {};
    for (let s in this.permutation.getAllStates()) {
        states[s] = this.permutation.getState(s);
    }
    return states;
}
Block.prototype.editState = function (state, id) {
    let states = {};
    for (let s in this.permutation.getAllStates()) {
        states[s] = this.permutation.getState(s);
    }
    for (let s in state) {
        states[s] = state[s];
    }
    try {
        this.setPermutation(BlockPermutation.resolve(id || this.typeId, states));
    } catch (err) { this.setType(id || this.typeId); throw err; }
}
BlockPermutation.prototype.edit = function (n) {
    let states = {};
    for (let s in this.getAllStates()) {
        states[s] = this.getState(s);
    }
    for (let s in n) {
        states[s] = n[s];
    }
    return BlockPermutation.resolve(this.type.id, states);
}





ItemStack.prototype.renameDynamicProperty = function (oldName, newName) {
    let tmp = this.getDynamicProperty(oldName);
    this.setDynamicProperty(oldName, null);
    this.setDynamicProperty(newName, tmp);
}
ItemStack.prototype.setAmount = function (n, add = true) {
    let total = this.amount;
    if (!isNaN(n)) {
        if (add) total += Number(n);
        else total = Number(n);
    }
    if (total <= 0) return null;
    this.amount = total;
    return this;
}





Entity.prototype.v3 = function () { return Vector.set(this.location); }
Entity.prototype.velocityV3 = function () { return Vector.set(this.getVelocity()); }
Entity.prototype.renameDynamicProperty = function (oldName, newName) {
    let tmp = this.getDynamicProperty(oldName);
    this.setDynamicProperty(oldName, null);
    this.setDynamicProperty(newName, tmp);
}
Entity.prototype.hasFamily = function (str) {
    let fam = this.getComponent('minecraft:type_family');
    if (fam) return fam.hasTypeFamily(str);
    return false;
}
Entity.prototype.getFamily = function () {
    let fam = this.getComponent('minecraft:type_family');
    if (fam) return fam.getTypeFamilies();
    return [];
}
Entity.prototype.isRideable = function () {
    return this.getComponent("minecraft:rideable") || null;
}
Entity.prototype.addRider = function (entity) {
    let ent = entity;
    let n = generateId(8);
    if (typeof entity === "string") this.runCommandAsync(`summon ${entity} ~~~ ~~ "" ${n}`).then(_ => {
        this.dimension.getEntities({ tags: [n], excludeTags: ["activated"], location: this.location, maxDistance: 1 }).some(e => {
            if (e.typeId.includes(entity)) ent = entity;
        });
    });
    return new Promise((res, rej) => { (ent && this.isRideable() && this.isRideable().addRider(ent)) ? res({ ride: this, rider: ent }) : rej({ ride: this, error: "Failed" }); });
}
Entity.prototype.getRider = function (id) {
    if (this.isRideable()) for (let e of this.isRideable().getRiders()) {
        if (e.id == id || e.typeId == id) return e;
    }
    return;
}
Entity.prototype.getRide = function () {
    try {
        if (!this.hasComponent('riding') || !this.getComponent('riding')) return null;
        return this.getComponent('riding').entityRidingOn;
    } catch (err) { }
    return null;
}
Entity.prototype.isRiding = function (id = null) {
    let ride = this.getRide();
    if (id) {
        if (ride && ride.isValid()) {
            if (ride.typeId === id || ride.id === id) return ride;
            if (typeof id === 'object') {
                for (let i = 0; i < id.length; i++) {
                    if (ride.typeId === id[i] || ride.id === id[i]) return ride;
                }
            }
        }
        return null
    }
    return ride;
}
Entity.prototype.getRiders = function () {
    if (!this.isRideable()) return [];
    return this.isRideable().getRiders();
}
Entity.prototype.getDriver = function () {
    let r;
    try {
        if (!this.isRideable()) return;
        r = this.isRideable().getRiders()[this.isRideable().controllingSeat];
    } catch (err) { }
    return r;
}
Entity.prototype.addDataAsTag = function (name, value) {
    for (let t of this.getTags()) {
        if (t.startsWith(`${name}=`)) this.removeTag(t);
    }
    this.addTag(`${name}=${value}`);
}
Entity.prototype.getDataFromTag = function (name) {
    for (let t of this.getTags()) {
        if (t.startsWith(`${name}=`)) return t.split('=')[1];
    }
    return null;
}
Entity.prototype.removeDataFromTag = function (name) {
    for (let t of this.getTags()) {
        if (t.startsWith(`${name}=`)) this.removeTag(t);
    }
}
Entity.prototype.addItem = function (item, amount = 1) {
    if (this.hasComponent('inventory')) {
        if (typeof item === "string") item = new ItemStack(item, amount);
        return this.getComponent('inventory').container.addItem(item);
    }
    return null;
}
Entity.prototype.getItem = function (slot) {
    let s = slot || this.selectedSlotIndex || 0;
    if (this.hasComponent('inventory')) return this.getComponent('inventory').container.getItem(s);
    return null;
}
Entity.prototype.setItem = function (item, slot) {
    if (!this.hasComponent('inventory')) return;
    let itemStack = item;
    if (itemStack) {
        if (typeof item === "string") {
            if (!item.includes(":")) itemStack = new ItemStack(`minecraft:${item}`);
            else itemStack = new ItemStack(item);
        }
        if (itemStack.typeId === "minecraft:air") itemStack = null;
        if (itemStack.amount == 0) itemStack = null;
    }
    let s = this.selectedSlotIndex;
    if (!isNaN(slot)) s = slot;
    this.getComponent('inventory').container.setItem(s, itemStack);
}
Entity.prototype.findItem = function (id) {
    if (!this.hasComponent('inventory')) return;
    let list = [], con = this.getComponent('inventory').container;
    for (let i = 0; i < con.size; i++) {
        if (con.getItem(i) && con.getItem(i).typeId.includes(id)) list.push({ item: con.getItem(i), slot: i });
    }
    return list;
}
Entity.prototype.editProperty = function (properties) {
    for (let i in properties) {
        try { this.setProperty(i, properties[i]);
        } catch (err) { throw `${err}\n${i}: ${properties[i]}`; }
    }
}
Entity.prototype.getBlock = function (x = 0, y = 0, z = 0) {
    return this.dimension.getBlock(this.v3().add(x, y, z)) || {};
}
Entity.prototype.setMarkVariant = function (n = 0) {
    this.getComponent('mark_variant').value = n;
    return this;
}
Entity.prototype.getMarkVariant = function () {
    return this.getComponent('mark_variant').value;
}
Entity.prototype.setSkinId = function (n=0) {
    this.getComponent('skin_id').value = n;
    return this;
}
Entity.prototype.getSkinId = function () {
    return this.getComponent('skin_id').value;
}





export function Vector(x = 0, y = x, z = x) {
    this.x = x;
    this.y = y;
    this.z = z;
}
Vector.prototype.toString = function () { return JSON.stringify({ ...this }); };
Vector.prototype.toLongString = function () { return `${this.x},${this.y},${this.z}`; };
Vector.set = function (x, y, z) {
    if (typeof x === "object") return new Vector(x.x, x.y, x.z);
    else if (!isNaN(x)) return new Vector(x, y, z);
    return new Vector();
};
Vector.fromString = function (str) {
    return Vector.set(JSON.parse(str));
}
Vector.fromLongString = function (str) {
    let s = str.split(",");
    return Vector.set(Number(s[0]), Number(s[1]), Number(s[2]));
}
Vector.prototype.set = function (x, y, z) {
    if (typeof x === "object") {
        this.x = x.x;
        this.y = x.y;
        this.z = x.z;
    } else if (!isNaN(x)) {
        this.x = x;
        this.y = [y, z, undefined].allEqual() ? x : y;
        this.z = [y, z, undefined].allEqual() ? x : z;
    }
    return this;
};
Vector.prototype.clone = function () { return new Vector(this.x, this.y, this.z); };
Vector.prototype.equal = function (x, y, z) {
    let v = Vector.set(x, y, z);
    return (this.x === v.x && this.y === v.y && this.z === v.z);
};
Vector.equal = function (a, b) { return Vector.set(a).equal(b); };
Vector.prototype.add = function (x, y, z) {
    if (typeof x === "object") {
        this.x += x.x;
        this.y += x.y;
        this.z += x.z;
    } else if (!isNaN(x)) {
        this.x += x;
        this.y += [y, z, undefined].allEqual() ? x : y;
        this.z += [y, z, undefined].allEqual() ? x : z;
    }
    return this;
};
Vector.add = function (v, x, y, z) { return Vector.set(v).add(x, y, z); };
Vector.prototype.sub = function (x, y, z) {
    if (typeof x === "object") {
        this.x -= x.x;
        this.y -= x.y;
        this.z -= x.z;
    } else if (!isNaN(x)) {
        this.x -= x;
        this.y -= [y, z, undefined].allEqual() ? x : y;
        this.z -= [y, z, undefined].allEqual() ? x : z;
    }
    return this;
};
Vector.sub = function (v, x, y, z) { return Vector.set(v).sub(x, y, z); };
Vector.prototype.mult = function (x, y, z) {
    if (typeof x === "object") {
        this.x *= x.x;
        this.y *= x.y;
        this.z *= x.z;
    } else if (!isNaN(x)) {
        this.x *= x;
        this.y *= [y, z, undefined].allEqual() ? x : y;
        this.z *= [y, z, undefined].allEqual() ? x : z;
    }
    return this;
};
Vector.mult = function (v, x, y, z) { return Vector.set(v).mult(x, y, z); };
Vector.prototype.divide = function (x, y, z) {
    if (typeof x === "object") {
        this.x /= x.x;
        this.y /= x.y;
        this.z /= x.z;
    } else if (!isNaN(x)) {
        this.x /= x;
        this.y /= [y, z, undefined].allEqual() ? x : y;
        this.z /= [y, z, undefined].allEqual() ? x : z;
    }
    return this;
};
Vector.divide = function (v, x, y, z) { return Vector.set(v).divide(x, y, z); };
Vector.prototype.getMag = function () { return Math.sqrt(this.x * this.x + this.y * this.y + this.z * this.z); };
Vector.prototype.getDirection = function (target) {
    let v = Vector.sub(target, this);
    let mag = v.getMag();
    return v.divide(mag, mag, mag);
};
Vector.prototype.getDistance = function (x, y, z) { return Vector.set(this).sub(x, y, z).getMag(); };
Vector.getDistance = function (v, a, b, c, d, e) {
    if (typeof v === "object") return Vector.set(v).sub(a, b, c).getMag();
    return new Vector(v, a, b).sub(c, d, e).getMag();
};
Vector.prototype.getYawPitch = function() {
    const yaw = -Math.atan2(this.x, this.z);
    const pitch = Math.atan2(this.y, Math.sqrt(this.x * this.x + this.z * this.z));
    const dist = this.getMag();
    return { yaw, pitch, dist };
}
Vector.applyAngle = function (yaw, pitch, magnitude = 1) {
    var deltaYawX = -Math.sin(yaw) * Math.cos(pitch) * magnitude;
    var deltaPitchY = Math.sin(pitch) * magnitude;
    var deltaYawZ = Math.cos(yaw) * Math.cos(pitch) * magnitude;
    return new Vector(deltaYawX, deltaPitchY, deltaYawZ);
}
Vector.prototype.ceil = function () { return new Vector(Math.ceil(this.x), Math.ceil(this.y), Math.ceil(this.z)); };
Vector.prototype.floor = function () { return new Vector(Math.floor(this.x), Math.floor(this.y), Math.floor(this.z)); };
Vector.prototype.round = function () { return new Vector(Math.round(this.x), Math.round(this.y), Math.round(this.z)); };
Vector.prototype.abs = function () { return new Vector(Math.abs(this.x), Math.abs(this.y), Math.abs(this.z)); };
Vector.prototype.above = function (n = 1) { const v = this.clone(); v.y += n; return v; };
Vector.prototype.below = function (n = 1) { const v = this.clone(); v.y -= n; return v; };
Vector.prototype.north = function (n = 1) { const v = this.clone(); v.z -= n; return v; };
Vector.prototype.south = function (n = 1) { const v = this.clone(); v.z += n; return v; };
Vector.prototype.east = function (n = 1) { const v = this.clone(); v.x += n; return v; };
Vector.prototype.west = function (n = 1) { const v = this.clone(); v.x -= n; return v; };






export const OVERWORLD = world.getDimension('overworld');
export const NETHER = world.getDimension('nether');
export const THE_END = world.getDimension('the_end');
world.renameDynamicProperty = function (oldName, newName) {
    let tmp = world.getDynamicProperty(oldName);
    world.setDynamicProperty(oldName, null);
    return world.setDynamicProperty(newName, tmp);
}
world.getPropertyIds = function () {
    let list = [];
    world.getDynamicPropertyIds().some(a => {
        if (a.endsWith("_part0")) list.push(a.replace(/_part0/g, ""));
    });
    return list;
}
world.setProperty = function (name, value, size = 32e3) {
    try {
        if (typeof value !== "string") value = JSON.stringify(value);
        let chunks = [];
        for (let i = 0; i < value.length; i += size) {
            chunks.push(value.slice(i, i + size));
        }
        for (let i = 0; i < chunks.length; i++) {
            world.setDynamicProperty(`${name}_part${i}`, chunks[i]);
        }
    } catch (err) { throw err }
}
world.getProperty = function (name = "") {
    try {
        let i = 0, current = world.getDynamicProperty(`${name}_part${i}`), data = "";
        while (current) {
            data += current;
            i += 1;
            current = world.getDynamicProperty(`${name}_part${i}`);
        }
        return data;
    } catch (err) { throw err }
}
world.getParsedProperty = function (id) {
    return JSON.parse(world.getProperty(id) || null);
}
world.clearProperty = function (name) {
    try {
        let i = 0, current = world.getDynamicProperty(`${name}_part${i}`);
        while (current) {
            world.setDynamicProperty(`${name}_part${i}`, null);
            i += 1;
            current = world.getDynamicProperty(`${name}_part${i}`);
        }
        return world;
    } catch (err) { throw err }
}
world.hasProperty = function (name) {
    for (let id of world.getDynamicPropertyIds()) {
        if (id.includes(name)) return true;
    }
    return false;
}
world.getPropertyByteTotal = function (name) {
    let dummy = new ItemStack("minecraft:wooden_hoe");
    let i = 0, current = world.getDynamicProperty(`${name}_part${i}`);
    while (current) {
        dummy.setDynamicProperty(`${name}_part${i}`, current);
        i += 1;
        current = world.getDynamicProperty(`${name}_part${i}`);
    }
    return dummy.getDynamicPropertyTotalByteCount();
}
world.clearStructure = function (id) {
    world.structureManager.getWorldStructureIds().forEach(a => {
        if (a.includes(id)) world.structureManager.delete(a);
    });
}
world.getStructure = function (id) {
    let list = [];
    world.structureManager.getWorldStructureIds().forEach(a => {
        if (a.includes(id)) list.push(world.structureManager.get(a));
    });
    return list;
}
world.createStructure = function (id, dimension, min, max, option) {
    let list = [];
    let size = Vector.sub(max, min);
    for (let z = 0; z <= Math.floor(size.z / 64); z++) {
        for (let x = 0; x <= Math.floor(size.x / 64); x++) {
            let _min = Vector.add(min, new Vector(x * 64, 0, z * 64));
            let _max = Vector.add(_min, new Vector(Math.min(64 - 1, size.x - (x * 64 - 1)), size.y, Math.min(64 - 1, size.z - (z * 64 - 1))));
            list.push(world.structureManager.createFromWorld(`${id}[${x},0,${z}]`, dimension, _min, _max, option));
        }
    }
    return list;
}
world.placeStructure = function (id, dimension, min, option) {
    world.getStructure(id).some(s => {
        let xz = s.id.match(/\[(.*?)\]/)[1];
        let offset = Vector.fromLongString(xz).mult(64);
        let pos = Vector.add(min, offset);
        world.structureManager.place(s.id, dimension, pos, option);
    });
}




export function generateId(length = 8, filter = []) {
    let COMBI = '';
    if (!filter.includes('ALPHABET')) {
        if (!filter.includes('UPPER_ALPHABET')) COMBI += ALPHABET;
        if (!filter.includes('LOWER_ALPHABET')) COMBI += ALPHABET.toLowerCase();
    }
    if (!filter.includes('NUMBERS')) COMBI += NUMBERS;
    if (!filter.includes('SPECIAL')) COMBI += SPECIAL;
    return ' '.repeat(length).replace(/ /g, _ => COMBI.charAt(Math.floor(Math.random() * COMBI.length)));
}
export function getEntity(id) {
    try {
        return world.getEntity(`${id}`);
    } catch (err) { }
    return null;
}
export function getDynamicPropertyByteCount(target, id) {
    let dummy = new ItemStack("minecraft:wooden_hoe");
    dummy.setDynamicProperty(id, target.getDynamicProperty(id));
    return dummy.getDynamicPropertyTotalByteCount();
}
export function angleDifference(a, b) {
    let diff = ((b - a + 180) % 360) - 180;
    return (diff < -180) ? (diff + 360) : diff;
}
export function formatBytes(a, b = 2) { if (!+a) return "0 Bytes"; const c = 0 > b ? 0 : b, d = Math.floor(Math.log(a) / Math.log(1024)); return `${parseFloat((a / Math.pow(1024, d)).toFixed(c))} ${["Bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"][d]}` }
export function map(n, start1, stop1, start2, stop2, within) {
    const newval = (n - start1) / (stop1 - start1) * (stop2 - start2) + start2;
    if (!within) return newval;
    if (start2 < stop2) return clamp(newval, start2, stop2);
    return clamp(newval, stop2, start2);
}
export function clamp(n, min, max) { return Math.min(Math.max(n, min), max); }
export default {
    _PI_180_,
    _180_PI_,
    HALF_PI,
    world,
    system,
    Vector,
    Player,
    Entity,
    BlockTypes,
    BlockPermutation,
    OVERWORLD, NETHER, THE_END,

    generateId,
    getEntity,
    getDynamicPropertyByteCount,
    angleDifference,
    formatBytes,
    map,
    clamp,

    ActionFormData, ModalFormData, uiManager
};