import { getEntity, system, Vector, world, _PI_180_ } from "./lib/Util";

export default function (ship) {
    const shipId = ship.id;
    const dimension = ship.dimension;
    const captain = getEntity(ship.getDataFromTag("captain"));
    const isSinking = ship.hasFamily("sinkingsheep");
    const bitik = dimension.spawnEntity("mship:bitik", ship.v3());
    const structures = {};

    const angle = Math.round(ship.currentYaw / 90) * 90;
    const data = world.getParsedProperty(ship.getDynamicProperty("mship:dataId"))
    if (!data || !ship.hasTag("mship:isActive")) return removeRelatedEntities(dimension, shipId, true);
    ship.removeTag("mship:isActive");

    if (captain && captain.isValid()) delete captain.hasMenuOpen;

    const shipPos = ship.v3().floor().above(data.y);
    const ang_diff = (360 + addAngles(angle, data.angle)) % 360;
    const rotation = { "0": "None", "90": "Rotate90", "180": "Rotate180", "270": "Rotate270" }[ang_diff];

    data.shipId = "";
    
    let struc = world.getStructure(data.id);
    struc.forEach((s, i) => {
        let xz = s.id.match(/\[(.*?)\]/)[1];
        let offset = Vector.fromLongString(xz).mult(64);
        let min = Vector.add(data.minPos, offset);
        let max = Vector.add(min, s.size);

        let strucPos = ship.v3().floor().below(data.lowestY - data.y);
        if (data.angle == 0) {
            if (ang_diff == 0) {
                strucPos.x -= (data.location.x - min.x);
                strucPos.z -= (data.location.z - min.z) - 1;
            }
            else if (ang_diff == 90) {
                strucPos.x -= (max.z - data.location.z) + 1;
                strucPos.z -= (data.location.x - min.x);
            }
            else if (ang_diff == 180) {
                strucPos.x -= (max.x - data.location.x);
                strucPos.z -= (max.z - data.location.z) + 1;
            }
            else if (ang_diff == 270) {
                strucPos.x -= (data.location.z - min.z) - 1;
                strucPos.z -= (max.x - data.location.x);
            }
        }
        else if (data.angle == 90) {
            if (ang_diff == 0) {
                strucPos.x -= (data.location.x - min.x) + 1;
                strucPos.z -= (data.location.z - min.z);
            }
            else if (ang_diff == 90) {
                strucPos.x -= (max.z - data.location.z);
                strucPos.z -= (data.location.x - min.x) + 1;
            }
            else if (ang_diff == 180) {
                strucPos.x -= (max.x - data.location.x) - 1;
                strucPos.z -= (max.z - data.location.z);
            }
            else if (ang_diff == 270) {
                strucPos.x -= (data.location.z - min.z);
                strucPos.z -= (max.x - data.location.x) - 1;
            }
        }
        else if (data.angle == -180) {
            if (ang_diff == 0) {
                strucPos.x -= (data.location.x - min.x);
                strucPos.z -= (data.location.z - min.z) + 1;
            }
            else if (ang_diff == 90) {
                strucPos.x -= (max.z - data.location.z) - 1;
                strucPos.z -= (data.location.x - min.x);
            }
            else if (ang_diff == 180) {
                strucPos.x -= (max.x - data.location.x);
                strucPos.z -= (max.z - data.location.z) - 1;
            }
            else if (ang_diff == 270) {
                strucPos.x -= (data.location.z - min.z) + 1;
                strucPos.z -= (max.x - data.location.x);
            }
        }
        else if (data.angle == -90) {
            if (ang_diff == 0) {
                strucPos.x -= (data.location.x - min.x) - 1;
                strucPos.z -= (data.location.z - min.z);
            }
            else if (ang_diff == 90) {
                strucPos.x -= (max.z - data.location.z);
                strucPos.z -= (data.location.x - min.x) - 1;
            }
            else if (ang_diff == 180) {
                strucPos.x -= (max.x - data.location.x) + 1;
                strucPos.z -= (max.z - data.location.z);
            }
            else if (ang_diff == 270) {
                strucPos.x -= (data.location.z - min.z);
                strucPos.z -= (max.x - data.location.x) + 1;
            }
        }
        structures[s.id] = { dimension, strucPos, rotation };
        if (i >= struc.length - 1) {
            placeStructures();
            system.runTimeout(finish, 2);
        }
    });

    function placeStructures(remove) {
        for (let i in structures) {
            const { dimension, strucPos, rotation } = structures[i];
            try {
                world.structureManager.place(i, dimension, strucPos, { rotation });
            } catch (err) { throw err; }
        }
        if (remove) world.clearStructure(data.id);
    }
    
    function finish() {
        let helm;
        try {
            removeRelatedEntities(dimension, shipId);

            findHelm:
            for (let z = -3; z < 3; z++) {
                for (let y = -3; y < 3; y++) {
                    for (let x = -3; x < 3; x++) {
                        let b = dimension.getBlock(Vector.add(shipPos, { x, y, z }));
                        if (b.hasTag("helm")) {
                            helm = b;
                            break findHelm;
                        }
                    }
                }
            }

            if (helm) {
                if (captain && captain.isValid()) {
                    let v = helm.v3().add(0.5 - Math.cos((angle + 90) * _PI_180_), 0.5, 0.5 - Math.sin((angle + 90) * _PI_180_));
                    if (!isSinking) captain.teleport(v, { rotation: { ...captain.getRotation(), y: angle } });
                    captain.sendMessage("Finishing up");
                }
                helm.editState({ "mship:used": 0 });
                data.location = helm.v3();
                data.facing = helm.permutation.getState('minecraft:cardinal_direction');
                data.angle = { 'west': -90, 'north': 0, 'east': 90, 'south': -180 }[data.facing];
                rotatePassengers(helm, data);
                rotateBlocks(helm, data);
                world.clearProperty(data.id);
                world.setProperty(data.id, JSON.stringify(data));
            }
            system.runTimeout(_ => placeStructures(true), 2);
        } catch (err) {
            console.warn(err);
        }

        if (bitik.isValid()) bitik.remove();
        if (ship.isValid()) ship.remove();
        if (!helm) {
            if (captain && captain.isValid()) captain.sendMessage("Error: Helm lost/missing".red());
            world.clearProperty(data.id);
        }
    }

    function rotatePassengers(helm, data) {
        for (let d of data.entities) {
            let e = getEntity(d.id);
            if (!(e && e.isValid())) continue;
            let v = Vector.add(translate(new Vector(), d.position, ang_diff), helm.bottomCenter()),
                b = dimension.getBlock(v), rotation = e.getRotation();
            rotation.y = (angle - d.baseAngle) + d.formerAngle + (e.formerYaw || 0);
            if (e.hasFamily("cannon")) v = b.bottomCenter();
            e.teleport(v, { rotation });
            if (e.getComponent("minecraft:mark_variant")) e.getComponent("minecraft:mark_variant").value = 0;
            e.setDynamicProperty("mship:data", null);
            e.removeTag("mship:isOnShip");
            delete e._yaw;
            delete e.riderYaw;
            delete e.formerYaw;
            delete e.turnTick;
        }
        data.entities = [];
    }

    function rotateBlocks(helm, data) {
        try {
            let tmp = {}, cache = [];
            data.minPos = helm.v3();
            data.maxPos = helm.v3();
            for (let i in data.blocks) {
                let t = translate(helm, data.blocks[i].position, ang_diff).floor();
                bitik.teleport(t);
                let b = dimension.getBlock(t);
                if (b && b.isValid()) {
                    if (b.typeId.includes("trapdoor")) {
                        let s = b.getState("direction") || 0, ns = s;
                        if (ang_diff == 90) {
                            if (s == 0) ns = [3, 0, 2, 1][Math.floor(ang_diff / 90)];
                            else if (s == 1) ns = [0, 2, 1, 3][Math.floor(ang_diff / 90)];
                            else if (s == 2) ns = [1, 3, 0, 2][Math.floor(ang_diff / 90)];
                            else if (s == 3) ns = [2, 1, 3, 0][Math.floor(ang_diff / 90)];
                        } else if (ang_diff == 180) {
                            if (s == 0) ns = [2, 1, 3, 0][Math.floor(ang_diff / 90)];
                            else if (s == 1) ns = [3, 0, 2, 1][Math.floor(ang_diff / 90)];
                            else if (s == 2) ns = [0, 2, 1, 3][Math.floor(ang_diff / 90)];
                            else if (s == 3) ns = [1, 3, 0, 2][Math.floor(ang_diff / 90)];
                        } else if (ang_diff == 270) {
                            if (s == 0) ns = [1, 3, 0, 2][Math.floor(ang_diff / 90)];
                            else if (s == 1) ns = [2, 1, 3, 0][Math.floor(ang_diff / 90)];
                            else if (s == 2) ns = [3, 0, 2, 1][Math.floor(ang_diff / 90)];
                            else if (s == 3) ns = [0, 2, 1, 3][Math.floor(ang_diff / 90)];
                        }
                        b.editState({ "direction": ns });
                    }
                }
                let v = Vector.sub(t, helm);
                data.minPos.set(Math.min(data.minPos.x, t.x), Math.min(data.minPos.y, t.y), Math.min(data.minPos.z, t.z));
                data.maxPos.set(Math.max(data.maxPos.x, t.x), Math.max(data.maxPos.y, t.y), Math.max(data.maxPos.z, t.z));
                tmp[v.toLongString()] = { position: v };
            }
            for (let c of data.cache) {
                cache.push(translate(new Vector(), Vector.fromLongString(c), ang_diff).toLongString());
            }
            data.blocks = tmp;
            data.cache = cache;
        } catch (err) { throw err; }
    }
}

function translate(base, location, angle) {
    let pos = Vector.add(base, location);
    if (angle == 90) pos = new Vector(base.x - location.z, base.y + location.y, base.z + location.x);
    else if (angle == -180 || angle == 180) pos = new Vector(base.x - location.x, base.y + location.y, base.z - location.z);
    else if (angle == 270) pos = new Vector(base.x + location.z, base.y + location.y, base.z - location.x);
    return pos;
}

function removeRelatedEntities(dimension, shipId) {
    dimension.getEntities({ tags: [`ship=${shipId}`] }).forEach(e => {
        e.removeTag("mship:isActive");
        e.removeTag(`ship=${shipId}`);
        if (e.hasFamily("mship") && (e.hasFamily("fragment") || e.hasFamily("fragment_collector") || e.hasFamily("ship") || e.hasFamily("sail_controller"))) e.remove();
    });
    dimension.runCommandAsync(`event entity @e[type=mship:bitik,tag="ship=${shipId}"] mship:despawn`);
}

function addAngles(a, b) {
    let n = ((a - b + 180) % 360) - 180;
    return (n < -180) ? (n + 360) : n;
}