import { world, system, Vector, _PI_180_ } from "./lib/Util";
import valid_blocks from "./valid_blocks";
import block_data from "./block_data";

export function scan(player, block, data) {
    if (!player || !block.hasTag("helm") || !data) return;
    let visited = [], callback = null, cancel = false;
    const counter = player.dimension.spawnEntity("mship:particle", block.bottomCenter());
    const max = Math.min(data.max, world.getDynamicProperty("mship:max_blocks"));
    const bitik = player.dimension.spawnEntity("mship:bitik", block.bottomCenter());
    counter.nameTag = `0`;

    data.spawnPoints = {};
    data.alerts = [];
    data.blocks = {};
    data.cache = [];
    data.dblocks = [];
    data.entities = [];
    data.y = 0;
    data.lowestY = 0;

    data.minPos = block.v3();
    data.maxPos = block.v3();

    function* loop(b) {
        if (callback) system.clearRun(callback);
        callback = system.runTimeout(finished, 2);
        if (!block.hasTag("helm")) return;
        if (Object.keys(data.blocks).length >= max) return data.alerts[0] = "Alert: Structure exceeds Max limit".yellow();
        if (visited.includes(b.toString())) return;
        if (data.valid[valid_blocks.indexOf(b.typeId)] != 1) return;

        let delta = Vector.sub(b, block);
        data.blocks[delta.toLongString()] = { position: delta };

        data.y = Math.max(data.y, -delta.y);
        data.lowestY = Math.max(data.y, -delta.y);
        data.minPos.set(Math.min(data.minPos.x, b.x), Math.min(data.minPos.y, b.y), Math.min(data.minPos.z, b.z));
        data.maxPos.set(Math.max(data.maxPos.x, b.x), Math.max(data.maxPos.y, b.y), Math.max(data.maxPos.z, b.z));
        bitik.teleport(b.v3());

        visited.push(b.toString());
        counter.nameTag = `Scanning\n[${visited.length}]`;
        for (let a of [b.above(), b.below(), b.north(), b.south(), b.west(), b.east()]) {
            if (!(a && a.isValid())) continue;
            if (a.typeId === "minecraft:air") continue;
            if (!valid_blocks.includes(a.typeId)) continue;
            if (data.valid[valid_blocks.indexOf(a.typeId)] != 1) continue;
            system.runJob(loop(a));
            yield;
        }
    }

    function finished() {
        if (!block.hasTag("helm")) return cancelled();
        block.editState({ "mship:used": 0 });

        world.getPlayers().some(ps => {
            try {
                if (ps.getSpawnPoint() == undefined) return;
                const base = ps.dimension.getBlock(ps.getSpawnPoint());
                if (base && base.isValid()) {
                    let b = base.below();
                    let v = Vector.sub(b, block);
                    if (!data.blocks[v.toLongString()]) return;
                    let p = v.clone();
                    p.distance = Math.sqrt(v.x * v.x + v.z * v.z);
                    p.angle = data.angle * _PI_180_ - Math.atan2(v.z, v.x);
                    data.spawnPoints[ps.id] = p;
                }
            } catch (err) { }
        });

        let msg = "Complete";
        if (data.alerts.length > 0) msg += `\n[\xa7e${data.alerts.length}\xa7r]`;
        counter.nameTag = msg;

        counter.triggerEvent("mship:complete");
        bitik.triggerEvent("mship:despawn");
        world.clearProperty(data.id).setProperty(data.id, data);
    }

    function cancelled() {
        world.clearProperty(data.id);
        counter.nameTag = "Cancelled";
        counter.triggerEvent("mship:complete");
        bitik.triggerEvent("mship:despawn");
        player.sendMessage(`Cancelled`);
    }
    system.runJob(loop(block));
}

export default { scan };