import { ActionFormData, ModalFormData, Block, clamp, Entity, formatBytes, ItemStack, Vector, world } from '../lib/Util';
import valid_blocks from '../valid_blocks';
import blockName from '../lib/blockName';
import { setupMenu } from './setup';
import compile from '../compile';
import { scan } from '../scan';

export function mainMenu(player, target, data) {
    if (!data) {
        if (target instanceof Entity) {
            if (target.hasMenuOpen) return;
            let id = target.getDynamicProperty("mship:dataId") || "";
            data = JSON.parse(world.getProperty(id) || null);
            target.hasMenuOpen = true;
        }
        else if (target instanceof Block) {
            world.getDynamicPropertyIds().some(a => {
                if (!a.endsWith("_part0")) return;
                data = JSON.parse(world.getProperty(a.replace("_part0", "")) || null);
                if (target.v3().equal(data.location)) return true;
                data = null;
            });
        }
    }
    if (!data) {
        exit();
        if (target instanceof Block) setupMenu(player, target);
        return;
    }
    if (target instanceof Block) {
        if (target.getState("mship:used")) return;
        target.editState({ "mship:used": 1 });
    }

    if (data.locked && player.id !== data.owner) {
        let item = player.getItem();
        if (item && (item.typeId === "mship:admin_key" || item.getDynamicProperty("mship:key") === data.id)) { }
        else {
            player.sendMessage(`Locked`.red());
            if (target instanceof Entity) delete target.hasMenuOpen;
            else if (target instanceof Block) delete target.editState({ "mship:used": 0 });
            return;
        }
    }

    const length = Object.keys(data.blocks).length;
    const totalBlocks = Math.max(length + data.cache.length, 0);
    const integrity = clamp(~~(100 * (length / totalBlocks)), 0, 100);
    const world_valid = world.getDynamicProperty("mship:valid_blocks");
    const min = Math.max(data.min, world.getDynamicProperty("mship:min_blocks"));
    const max = Math.min(data.max, world.getDynamicProperty("mship:max_blocks"));
    const controller = world.getDynamicProperty("mship:controller");
    const totalSize = world.getPropertyByteTotal(data.id);
    let errors = "", alerts = data.alerts;

    let integrityColor = '\xA7a';
    if (integrity < 74) integrityColor = '\xA7e';
    if (integrity < 40) integrityColor = '\xA7c';

    if (integrity < 40) errors += `Error: Broken\n`;
    if (target instanceof Block) {
        let hasHelm = false
        for (let i in data.blocks) {
            let v = data.blocks[i].position;
            let b = target.offset(v);
            if (b && b.typeId.includes("_helm")) {
                if (hasHelm) {
                    errors += `Error: Too many helms connected\n`;
                    break;
                }
                else hasHelm = true;
            }
        }
    }
    if (length < min || length > max) {
        errors += `Error: Connected blocks must be within ${min} and ${max}\n`;
    }

    function main() {
        const raw = { rawtext: [] };
        raw.rawtext.push({ text: `Name: ${data.name}\n` });
        raw.rawtext.push({ text: `Integrity: ${integrityColor}${integrity}%% `.reset() + `[${length.toString().green()} / ${totalBlocks.toString().green()}] \n`.reset() });
        raw.rawtext.push({ text: `Facing: ${data.facing.toString().capFirstLetter()} [${data.angle.toString().green()}\u00b0]\n` });
        raw.rawtext.push({ text: "\n" });
        raw.rawtext.push({ text: `Blocks: ${length.toString().green()} / ${max.toString().green()}\n` });
        raw.rawtext.push({ text: "\n" });
        raw.rawtext.push({ text: `Locked: ${data.locked.toString().blue()}\n` });
        raw.rawtext.push({ text: `Data size: ${formatBytes(totalSize).green()}\n\n` });
        if (alerts) raw.rawtext.push({ text: `\xa7e${alerts}\xa7r\n` });
        raw.rawtext.push({ text: `\xa7c${errors}\xa7r\n` });

        const form = new ActionFormData().title(`Ship Menu`);
        form.button("Configure");
        if (errors === "" && target instanceof Block) form.button("Compile".dark_purple().strong());
        else if (target instanceof Entity) form.button("Align");
        form.body(raw);
        form.show(player).then(res => {
            if (res.cancelationReason == "UserBusy") return;
            else if (res.selection == 0) configure();
            else if (res.selection == 1) {
                if (target instanceof Block) compile(player, target, data);
                else align(target);
            }
            else exit();
        }).catch(exit);
    }

    function configure() {
        let form = new ActionFormData().title(`Configure`);
        if (target instanceof Entity) {
            form.button('Camera');
            if (controller) form.button(`Get Controller`);
        }
        else {
            form.button('Settings');
            form.button(`Scan`.dark_green().strong());
            form.button(`Create Key`);
            form.button(`Create Compass`);
            form.button('Valid blocks');
        }
        form.show(player).then(res => {
            if (target instanceof Entity) {
                if (res.selection == 0) camera();
                else if (res.selection == 1) {
                    if (controller) {
                        if (player.findItem("mship:controller").length == 0) player.addItem("mship:controller");
                    }
                    delete target.hasMenuOpen;
                }
                else return recall(player, target, data);
            } else {
                if (res.selection == 0) settings();
                else if (res.selection == 1) scan(player, target, data);
                else if (res.selection == 2) createKey();
                else if (res.selection == 3) createCompass();
                else if (res.selection == 4) validBlocks();
                else return recall(player, target, data);
            }
        }).catch(exit);
    }

    function camera() {
        new ModalFormData().title("Configure Camera")
            .toggle("Camera", data.camera)
            .slider("Camera Distance 10~32\n", 5, 32, 1, data.cameraDistance)
            .show(player).then(res => {
                data.camera = res.formValues[0];
                data.cameraDistance = res.formValues[1];
                world.setProperty(data.id, JSON.stringify(data));
                recall(player, target, data);
            }).catch(configure);
    }

    function settings() {
        new ModalFormData().title("Configure Settings")
            .textField("Name", "Input name", `${data.name}`)
            .toggle('\u00A7rLock', data.locked)
            .slider(`\n\u00A7c\u00A7o* Higher Values may cause lag *\u00A7r\nMaximum Blocks`, world.getDynamicProperty("mship:min_blocks"), world.getDynamicProperty("mship:max_blocks"), 1, max)
            .show(player).then(res => {
                data.name = `${res.formValues[0] || `${player.nameTag}' Ship`}`;
                data.locked = res.formValues[1];
                if (data.max != Number(res.formValues[2])) changes({ "max": Number(res.formValues[2]) });
                else {
                    world.setProperty(data.id, JSON.stringify(data));
                    recall(player, target, data);
                }
            }).catch(configure);
    }

    function createKey() {
        const raw = { rawtext: [] };
        raw.rawtext.push({ text: "An intricately designed key that unlocks the ship's helm, granting access to its control menu." });
        raw.rawtext.push({ text: "\n" });
        raw.rawtext.push({ text: "\n" });
        raw.rawtext.push({ text: "Cost(s)\n\n" });
        raw.rawtext.push({ text: "\ue110 x1 " });
        raw.rawtext.push({ text: "\n\n\n\n\n\n\n" });
        new ActionFormData().title('Key Creation').body(raw).button('Create')
            .show(player).then(res => {
                if (res.selection == 0) {
                    let key = new ItemStack("mship:key");
                    key.nameTag = `${data.name}`.dark_purple();
                    key.setDynamicProperty("mship:key", data.id);
                    if (player.getGameMode() === "survival") {
                        let list = player.findItem("minecraft:gold_ingot");
                        if (list.length > 0) {
                            player.setItem(list[0].item.setAmount(-1), list[0].slot);
                            player.addItem(key);
                            player.sendMessage("Key created");
                        } else player.sendMessage(`Key creation failed`.red());
                    } else {
                        player.addItem(key);
                        player.sendMessage("Key created");
                    }
                    if (target instanceof Block) target.editState({ "mship:used": 0 });
                } else recall(player, target, data);
            }).catch(configure);
    }

    function createCompass() {
        const raw = { rawtext: [] };
        raw.rawtext.push({ text: "A magical compass bound to a specific ship, always guiding its owner back to the vessel, no matter the distance or location." });
        raw.rawtext.push({ text: "\n" });
        raw.rawtext.push({ text: "\n" });
        raw.rawtext.push({ text: "Cost(s)\n\n" });
        raw.rawtext.push({ text: "\ue108 x1 " });
        raw.rawtext.push({ text: "\ue109 x1 " });
        raw.rawtext.push({ text: "\ue110 x4 " });
        raw.rawtext.push({ text: "\n\n\n\n\n\n" });
        new ActionFormData().title('Compass Creation').body(raw).button('Create')
            .show(player).then(res => {
                if (res.selection == 0) {
                    const item = new ItemStack("mship:ship_compass0");
                    item.nameTag = `${data.name} Compass`.dark_purple();
                    item.setDynamicProperty("mship:dataId", data.id);
                    if (player.getGameMode() === "survival") {
                        const gold = player.findItem("minecraft:gold_ingot");
                        const compass = player.findItem("minecraft:compass");
                        if (gold.length > 0 && compass.length > 0) {
                            player.setItem(gold[0].item.setAmount(-1), gold[0].slot);
                            player.setItem(compass[0].item.setAmount(-1), compass[0].slot);
                            player.addItem(item);
                            player.sendMessage("Compass created");
                        } else player.sendMessage(`Compass creation failed`.red() + "\nNot enough resources");
                    } else {
                        player.addItem(item);
                        player.sendMessage("Compass created");
                    }
                    if (target instanceof Block) target.editState({ "mship:used": 0 });
                } else recall(player, target, data);
            }).catch(configure);
    }

    function validBlocks() {
        const modal = new ModalFormData().title("Configure Valid blocks");
        player.onScreenDisplay.setActionBar("Loading...");
        for (let i = 0; i < data.valid.length; i++) {
            let id = valid_blocks[i];
            if (id.includes('_helm') || id.includes("double") || id.includes("lit_") || id.includes("_anvil") || id.includes("sponge")) continue;
            let name = blockName(id);
            if (world_valid[i] == 0) continue;
            modal.toggle(name, data.valid[i] == 1 && world_valid[i] == 1);
        }
        modal.show(player).then(res => {
            let dup = `${world_valid}`, index = 0;
            for (let i = 0; i < data.valid.length; i++) {
                let id = valid_blocks[i];
                if (id.includes('_helm') || id.includes("double") || id.includes("lit_") || id.includes("_anvil") || id.includes("sponge")) continue;
                if (world_valid[i] == 0) continue;
                let tmp = (res.formValues[index++] == true) ? '1' : '0';
                dup = dup.substring(0, i) + tmp + dup.substring(i + 1);
            }
            if (dup !== data.valid) changes({ 'valid': dup });
        }).catch(configure);
    }

    function changes(n) {
        let form = new ActionFormData().title('Changes notification');
        let raw = { rawtext: [{ text: "Changes found:\n" }] };
        for (let i in n) {
            if (i === "max") raw.rawtext.push({ text: `Maximum blocks: ${data.max.toString().green()} => ${n[i].toString().green()}\n` });
            else if (i === "valid") {
                raw.rawtext.push({ text: "Valid blocks:\n" });
                for (let l = 0; l < world_valid.length; l++) {
                    if ([data.valid[l]] != n[i][l]) {
                        let name = blockName(valid_blocks[l]);
                        if (n[i][l] == 1) raw.rawtext.push({ rawtext: [{ text: "\u00A7a + " }, name, { text: "\u00A7r\n" }] });
                        else raw.rawtext.push({ rawtext: [{ text: "\u00A7c - " }, name, { text: "\u00A7r\n" }] });
                    }
                }
            }
        }
        let lb = Math.max(10 - raw.rawtext.length, 0);
        raw.rawtext.push({ text: "\n".repeat(lb) });
        form.body(raw);
        form.button("Confirm & Rescan");
        form.button('Cancel');
        form.show(player).then(res => {
            if (res.selection == 0) {
                for (let i in n) {
                    data[i] = n[i];
                }
                world.setProperty(data.id, JSON.stringify(data));
                if (target instanceof Block) scan(player, target, data);
            }
            else if (res.selection == 1) return configure();
            else if (res.canceled) configure();
        }).catch(configure);
    }

    function exit() {
        if (target instanceof Block) target.editState({ "mship:used": 0 });
        else if (target instanceof Entity) delete target.hasMenuOpen;
    }

    function recall(player, target, data) {
        exit();
        mainMenu(player, target, data);
    }

    main();
}

function align(ship) {
    let angle = Math.round(ship.getRotation().y / 90) * 90;
    /*
    ship.addTag("mship:toRotate");
    ship.headingError = 0;
    ship.currentYaw = angle;
    player.runCommandAsync(`tp @s ~~~ ${angle}~`)
        .then(_ => ship.runCommandAsync(`tp @s ~~0.185~ ${angle}~`))
        .then(_ => ship.addRider(player))
        .then(_ => ship.removeTag("mship:toRotate"));
        */
    ship.currentYaw = angle;
    delete ship.hasMenuOpen;
}

export default { mainMenu }