import { ActionFormData, formatBytes, ModalFormData, system, world } from "../lib/Util";
import blockName from "../lib/blockName";
import valid_blocks from "../valid_blocks";


export function globalMenu(player) {
    let valid = world.getDynamicProperty("mship:valid_blocks");
    let min = world.getDynamicProperty("mship:min_blocks");
    let max = world.getDynamicProperty("mship:max_blocks");
    let land_speed = world.getDynamicProperty("mship:land_speed");
    let water_speed = world.getDynamicProperty("mship:water_speed");
    let controller = world.getDynamicProperty("mship:controller");
    let totalShips = 0;
    let totalShipsActive = 0;
    let totalDataSize = world.getDynamicPropertyTotalByteCount();

    totalShips = world.getDynamicPropertyIds().filter(a => (a.startsWith("mship:ship") && !a.includes("|"))).length;
    totalShipsActive = player.dimension.getEntities({ type: "mship:ship" }).length;

    function main(err) {
        if (err) throw err;
        let body = '', lb = 12;
        body += `Minimum blocks: ${min.toString().green()}\n`;
        body += `Maximum blocks: ${max.toString().green()}\n`;
        body += `\n`;
        body += `Speed on water: ${water_speed.toString().green()}\n`;
        body += `Speed on land: ${land_speed.toString().green()}\n`
        body += `\n`;
        body += `Total ships: ${totalShips.toString().green()}\n`;
        body += `Active ships: ${totalShipsActive.toString().green()}\n`;
        body += `\n`
        body += `Controller: ${controller.toString().blue()}`;
        body += `\n`
        body += `Total Data size: ${formatBytes(totalDataSize).green()}`;
        body += `\n`;

        new ActionFormData().title(`World Settings`)
            .body(body + '\n'.repeat(lb - body.match(/\n/g).length))
            .button("Configure")
            .show(player).then(res => {
                if (res.cancelationReason == "UserBusy") system.run(main);
                if (res.selection === 0) configure();
            });
    }

    function configure() {
        new ActionFormData().title(`Configure`)
            .button("Settings")
            .button("Valid blocks")
            .button("Clear Data".dark_red())
            .show(player).then(res => {
                if (res.selection === 0) settings();
                else if (res.selection === 1) validBlocks();
                else if (res.selection === 2) clearDataMenu();
                else if (res.canceled) globalMenu(player);
            });
    }

    function settings() {
        let sliderMaxCap = 5124;
        new ModalFormData().title(`Settings`)
            .slider(`\nMinimum blocks cap\xA7a`, 12, sliderMaxCap, 1, min)
            .slider(`\xA7rMaximum blocks cap\xA7a`, 12, sliderMaxCap, 1, max)
            .slider(`\xA7r\nSpeed on water\n10 = 1 block/s\xA7a`, 10, 200, 1, water_speed)
            .slider(`\xA7rSpeed on land\n10 = 1 block/s\xA7a`, 10, 200, 1, land_speed)
            .toggle("Controller", controller)
            .show(player).then(res => {
                let n = {}, val = res.formValues;
                val[1] = Math.min(Math.max(val[0] + 12, val[1]), sliderMaxCap);
                val[0] = Math.max(Math.min(val[0], val[1] - 12), 12);
                if (min != val[0]) n["mship:min_blocks"] = val[0];
                if (max != val[1]) n["mship:max_blocks"] = val[1];
                if (water_speed !== val[2]) n["mship:water_speed"] = val[2];
                if (land_speed !== val[3]) n["mship:land_speed"] = val[3];
                if (controller !== val[4]) n["mship:controller"] = val[4];
                if (Object.keys(n).length > 0) changes(n);
            }).catch(configure);
    }

    function validBlocks() {
        const modal = new ModalFormData().title("Configure Valid blocks");
        player.onScreenDisplay.setActionBar("Loading...");
        for (let i = 0; i < valid.length; i++) {
            let id = valid_blocks[i];
            if (id.includes('_helm') || id.includes("double") || id.includes("lit_") || id.includes("_anvil") || id.includes("sponge")) continue;
            let name = blockName(id);
            modal.toggle(name, valid[i] == 1);
        }
        modal.show(player).then(res => {
            let dup = `${valid}`, index = 0;
            for (let i = 0; i < valid.length; i++) {
                let id = valid_blocks[i];
                if (id.includes('_helm') || id.includes("double") || id.includes("lit_") || id.includes("_anvil") || id.includes("sponge")) continue;
                let tmp = (res.formValues[index++] == true) ? '1' : '0';
                dup = dup.substring(0, i) + tmp + dup.substring(i + 1);
            }
            if (dup !== valid) changes({ 'mship:valid_blocks': dup });
        }).catch(configure);
    }

    function clearDataMenu() {
        let body = "Permanently delete all ships data in the world.".red();
        new ActionFormData().title('Clearing Data')
            .body(body + '\n'.repeat(11) + `Make sure all ships are deactivated.\n`.red().italic())
            .button('Continue')
            .show(player).then(res => {
                if (res.selection == 0) clearWorldData(player);
                if (res.canceled) configure();
            }).catch(e => { throw e });
    }

    function changes(n) {
        let form = new ActionFormData().title('Changes notification');
        let raw = { rawtext: [{ text: "Changes found:\n" }] };
        for (let i in n) {
            if (i === 'mship:min_blocks') raw.rawtext.push({ text: `Minimum blocks: ${min.toString().green()} => ${n[i].toString().green()}\n` });
            else if (i === 'mship:max_blocks') raw.rawtext.push({ text: `Maximum blocks: ${max.toString().green()} => ${n[i].toString().green()}\n` });
            else if (i === 'mship:water_speed') raw.rawtext.push({ text: `Speed on water: ${water_speed.toString().green()} => ${n[i].toString().green()}\n` });
            else if (i === 'mship:land_speed') raw.rawtext.push({ text: `Speed on land: ${land_speed.toString().green()} => ${n[i].toString().green()}\n` });
            else if (i === 'mship:controller') raw.rawtext.push({ text: `Controller: ${controller.toString().blue()} => ${n[i].toString().blue()}\n` });
            else if (i === 'mship:valid_blocks') {
                raw.rawtext.push({ text: "Valid blocks:\n" });
                for (let l = 0; l < valid.length; l++) {
                    if ([valid[l]] != n[i][l]) {
                        let name = blockName(valid_blocks[l]);
                        if (n[i][l] == 1) raw.rawtext.push({ rawtext: [{ text: "\u00A7a + " }, name, { text: "\u00A7r\n" }] });
                        else raw.rawtext.push({ rawtext: [{ text: "\u00A7c - " }, name, { text: "\u00A7r\n" }] });
                    }
                }
            }
        }
        let note = `\n* Changes will apply to all ships *\n`.red().italic();
        let lb = Math.max(5 - raw.rawtext.length, 0);
        raw.rawtext.push({ text: "\n".repeat(lb) });
        raw.rawtext.push({ text: note });
        form.body(raw);
        form.button("Confirm & Update");
        form.button('Cancel');

        form.show(player).then(res => {
            if (res.selection == 0) {
                for (let i in n) {
                    world.setDynamicProperty(i, n[i]);
                    if (i === 'mship:valid_blocks') {
                        world.getDynamicPropertyIds().some(a => {
                            if (!a.endsWith("_part0")) return;
                            const data = world.getParsedProperty(a.replace("_part0", ""));
                            if (!data) return;
                            let v = "";
                            for (let j = 0; j < n[i].length; j++) {
                                if (n[i][j] == 0) v += "0";
                                else v += data.valid.charAt(j);
                            }
                            data.valid = v;
                            world.clearProperty(data.id).setProperty(data.id, data);
                        });
                    }
                }
                system.runTimeout(_ => globalMenu(player), 20);
            }
            else if (res.selection == 1) return configure();
            else if (res.canceled) configure();
        }).catch(main);
    }

    main();
}

export function clearWorldData(player) {
    try {
        let totalDatas = 0;
        world.getDynamicPropertyIds().forEach(a => {
            if (!a.startsWith("mship")) return;
            world.setDynamicProperty(a, null);
            totalDatas += 1;
        });
        world.structureManager.getWorldStructureIds().forEach(a => {
            if (!a.startsWith("mship")) return;
            world.structureManager.delete(a);
        });
        player.dimension.runCommandAsync(`event entity @e mship:despawn`);
        player.dimension.getEntities({ families: ["mship"] }).forEach(ent => ent.remove());
        player.sendMessage(`${totalDatas.toString().green()} Data Cleared`.red());
    } catch (err) { throw err; }
    return player;
}



export default { globalMenu, clearWorldData }